from opentera.db.Base import db, BaseModel


class TeraDeviceSubType(db.Model, BaseModel):

    __tablename__ = 't_devices_subtypes'
    id_device_subtype = db.Column(db.Integer, db.Sequence('id_device_subtype_sequence'), primary_key=True,
                                  autoincrement=True)
    id_device_type = db.Column(db.Integer, db.ForeignKey('t_devices_types.id_device_type', ondelete='cascade'),
                               nullable=False)
    device_subtype_name = db.Column(db.String, nullable=False)

    device_subtype_type = db.relationship("TeraDeviceType")

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields += ['device_subtype_type']

        device_subtype_json = super().to_json(ignore_fields=ignore_fields)

        # Add device type name as parent
        device_subtype_json['device_subtype_parent'] = self.device_subtype_type.device_type_name

        return device_subtype_json

    @staticmethod
    def create_defaults(test=False):
        from opentera.db.models.TeraDeviceType import TeraDeviceType
        bureau = TeraDeviceType.get_device_type_by_key('bureau_actif')
        subtype = TeraDeviceSubType()
        subtype.device_subtype_type = bureau
        subtype.device_subtype_name = 'Bureau modèle #1'
        db.session.add(subtype)

        subtype = TeraDeviceSubType()
        subtype.device_subtype_type = bureau
        subtype.device_subtype_name = 'Bureau modèle #2'
        db.session.add(subtype)

        db.session.commit()

    @staticmethod
    def get_devices_subtypes():
        return TeraDeviceSubType.query.all()

    @staticmethod
    def get_device_subtype(dev_subtype: int):
        return TeraDeviceSubType.query.filter_by(id_device_subtype=dev_subtype).first()

    @staticmethod
    def get_device_subtypes_for_type(dev_type: int):
        return TeraDeviceSubType.query.filter_by(id_device_type=dev_type).all()




from opentera.db.Base import db, BaseModel


class TeraServiceProject(db.Model, BaseModel):
    __tablename__ = 't_services_projects'
    id_service_project = db.Column(db.Integer, db.Sequence('id_service_project_sequence'), primary_key=True,
                                   autoincrement=True)
    id_service = db.Column(db.Integer, db.ForeignKey('t_services.id_service', ondelete='cascade'), nullable=False)
    id_project = db.Column(db.Integer, db.ForeignKey('t_projects.id_project', ondelete='cascade'), nullable=False)

    service_project_service = db.relationship("TeraService")
    service_project_project = db.relationship("TeraProject")

    def __init__(self):
        pass

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_project_service', 'service_project_project'])

        if minimal:
            ignore_fields.extend([])

        return super().to_json(ignore_fields=ignore_fields)

    @staticmethod
    def get_services_for_project(id_project: int):
        return TeraServiceProject.query.filter_by(id_project=id_project).all()

    @staticmethod
    def get_projects_for_service(id_service: int):
        return TeraServiceProject.query.filter_by(id_service=id_service).all()

    @staticmethod
    def get_service_project_by_id(service_project_id: int):
        return TeraServiceProject.query.filter_by(id_service_project=service_project_id).first()

    @staticmethod
    def get_service_project_for_service_project(project_id: int, service_id: int):
        return TeraServiceProject.query.filter_by(id_project=project_id, id_service=service_id).first()

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraService import TeraService
            from opentera.db.models.TeraProject import TeraProject

            project1 = TeraProject.get_project_by_projectname('Default Project #1')
            project2 = TeraProject.get_project_by_projectname('Default Project #2')

            servicebureau = TeraService.get_service_by_key('BureauActif')
            # serviceviddispatch = TeraService.get_service_by_key('VideoDispatch')
            servicevideorehab = TeraService.get_service_by_key('VideoRehabService')

            service_project = TeraServiceProject()
            service_project.id_project = project1.id_project
            service_project.id_service = servicebureau.id_service
            db.session.add(service_project)

            service_project = TeraServiceProject()
            service_project.id_project = project1.id_project
            service_project.id_service = servicevideorehab.id_service
            db.session.add(service_project)

            service_project = TeraServiceProject()
            service_project.id_project = project2.id_project
            service_project.id_service = servicebureau.id_service
            db.session.add(service_project)

            db.session.commit()

    @staticmethod
    def delete_with_ids(service_id: int, project_id: int):
        delete_obj = TeraServiceProject.query.filter_by(id_service=service_id, id_project=project_id).first()
        if delete_obj:
            db.session.delete(delete_obj)
            db.session.commit()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: JoinSessionEvent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='JoinSessionEvent.proto',
  package='opentera.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16JoinSessionEvent.proto\x12\x11opentera.protobuf\"\xed\x01\n\x10JoinSessionEvent\x12\x13\n\x0bsession_url\x18\x01 \x01(\t\x12\x1c\n\x14session_creator_name\x18\x02 \x01(\t\x12\x14\n\x0csession_uuid\x18\x03 \x01(\t\x12\x1c\n\x14session_participants\x18\x04 \x03(\t\x12\x15\n\rsession_users\x18\x05 \x03(\t\x12\x17\n\x0fsession_devices\x18\x06 \x03(\t\x12\x10\n\x08join_msg\x18\x07 \x01(\t\x12\x1a\n\x12session_parameters\x18\x08 \x01(\t\x12\x14\n\x0cservice_uuid\x18\t \x01(\tb\x06proto3'
)




_JOINSESSIONEVENT = _descriptor.Descriptor(
  name='JoinSessionEvent',
  full_name='opentera.protobuf.JoinSessionEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_url', full_name='opentera.protobuf.JoinSessionEvent.session_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_creator_name', full_name='opentera.protobuf.JoinSessionEvent.session_creator_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_uuid', full_name='opentera.protobuf.JoinSessionEvent.session_uuid', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_participants', full_name='opentera.protobuf.JoinSessionEvent.session_participants', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_users', full_name='opentera.protobuf.JoinSessionEvent.session_users', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_devices', full_name='opentera.protobuf.JoinSessionEvent.session_devices', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='join_msg', full_name='opentera.protobuf.JoinSessionEvent.join_msg', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='session_parameters', full_name='opentera.protobuf.JoinSessionEvent.session_parameters', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_uuid', full_name='opentera.protobuf.JoinSessionEvent.service_uuid', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=46,
  serialized_end=283,
)

DESCRIPTOR.message_types_by_name['JoinSessionEvent'] = _JOINSESSIONEVENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

JoinSessionEvent = _reflection.GeneratedProtocolMessageType('JoinSessionEvent', (_message.Message,), {
  'DESCRIPTOR' : _JOINSESSIONEVENT,
  '__module__' : 'JoinSessionEvent_pb2'
  # @@protoc_insertion_point(class_scope:opentera.protobuf.JoinSessionEvent)
  })
_sym_db.RegisterMessage(JoinSessionEvent)


# @@protoc_insertion_point(module_scope)



from opentera.db.Base import db, BaseModel


class TeraServiceRole(db.Model, BaseModel):
    __tablename__ = 't_services_roles'
    id_service_role = db.Column(db.Integer, db.Sequence('id_service_role_sequence'), primary_key=True,
                                autoincrement=True)
    id_service = db.Column(db.Integer, db.ForeignKey('t_services.id_service', ondelete='cascade'), nullable=False)
    # Specific project role for a project, used mostly with OpenTera service for project access
    id_project = db.Column(db.Integer, db.ForeignKey('t_projects.id_project', ondelete='cascade'), nullable=True)
    # Specific site role for a site, used mostly with OpenTera service for site access
    id_site = db.Column(db.Integer, db.ForeignKey('t_sites.id_site', ondelete='cascade'), nullable=True)
    service_role_name = db.Column(db.String(100), nullable=False)

    service_role_service = db.relationship("TeraService")
    service_role_project = db.relationship('TeraProject')
    service_role_site = db.relationship('TeraSite')

    def __init__(self):
        pass

    def __str__(self):
        return '<TeraServiceRole Service = ' + str(self.service_role_service.service_name) + ', Role = ' + \
               str(self.service_role_name) + ' >'

    def __repr__(self):
        return self.__str__()

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_role_service', 'service_role_project', 'service_role_site'])

        if minimal:
            ignore_fields.extend([])

        json_val = super().to_json(ignore_fields=ignore_fields)

        # Remove null values
        if not json_val['id_project']:
            del json_val['id_project']
        if not json_val['id_site']:
            del json_val['id_site']
        else:
            if not minimal:
                json_val['site_name'] = self.service_role_site.site_name

        return json_val

    @staticmethod
    def get_service_roles(service_id: int):
        return TeraServiceRole.query.filter_by(id_service=service_id).all()

    @staticmethod
    def get_service_roles_for_site(service_id: int, site_id: int):
        return TeraServiceRole.query.filter_by(id_service=service_id, id_site=site_id).all()

    @staticmethod
    def get_specific_service_role_for_site(service_id: int, site_id: int, rolename: str):
        return TeraServiceRole.query.filter_by(id_service=service_id, id_site=site_id, service_role_name=rolename)\
            .first()

    @staticmethod
    def get_service_roles_for_project(service_id: int, project_id: int):
        return TeraServiceRole.query.filter_by(id_service=service_id, id_project=project_id).all()

    @staticmethod
    def get_specific_service_role_for_project(service_id: int, project_id: int, rolename: str):
        return TeraServiceRole.query.filter_by(id_service=service_id, id_project=project_id,
                                               service_role_name=rolename).first()

    @staticmethod
    def get_service_role_by_id(role_id: int):
        return TeraServiceRole.query.filter_by(id_service_role=role_id).first()

    @staticmethod
    def create_defaults(test=False):
        from opentera.db.models.TeraService import TeraService

        for service in TeraService.query.all():
            if service.service_key != 'OpenTeraServer':  # Don't add global roles for TeraServer
                new_role = TeraServiceRole()
                new_role.id_service = service.id_service
                new_role.service_role_name = 'admin'
                db.session.add(new_role)

                new_role = TeraServiceRole()
                new_role.id_service = service.id_service
                new_role.service_role_name = 'user'
                db.session.add(new_role)
            else:
                pass  # TODO: do what we did in Project and Site Access

        db.session.commit()



from opentera.db.Base import db, BaseModel


class TeraSessionUsers(db.Model, BaseModel):
    __tablename__ = 't_sessions_users'
    id_session_user = db.Column(db.Integer, db.Sequence('id_session_user'), primary_key=True, autoincrement=True)
    id_session = db.Column(db.Integer, db.ForeignKey('t_sessions.id_session', ondelete='cascade'))
    id_user = db.Column(db.Integer, db.ForeignKey('t_users.id_user'))

    session_user_session = db.relationship('TeraSession', cascade="delete")
    session_user_user = db.relationship('TeraUser')



from opentera.db.Base import db, BaseModel


class TeraDeviceParticipant(db.Model, BaseModel):
    __tablename__ = 't_devices_participants'
    id_device_participant = db.Column(db.Integer, db.Sequence('id_device_participant_sequence'), primary_key=True,
                                      autoincrement=True)
    id_device = db.Column(db.Integer, db.ForeignKey("t_devices.id_device"), nullable=False)
    id_participant = db.Column(db.Integer, db.ForeignKey("t_participants.id_participant", ondelete='cascade'),
                               nullable=False)

    device_participant_participant = db.relationship("TeraParticipant")
    device_participant_device = db.relationship("TeraDevice")

    def to_json(self, ignore_fields=[], minimal=False):
        ignore_fields.extend(['device_participant_participant', 'device_participant_device'])

        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraParticipant import TeraParticipant
            from opentera.db.models.TeraDevice import TeraDevice
            participant1 = TeraParticipant.get_participant_by_id(1)
            participant2 = TeraParticipant.get_participant_by_id(2)
            device1 = TeraDevice.get_device_by_name('Apple Watch #W05P1')
            device2 = TeraDevice.get_device_by_name('Kit Télé #1')
            device3 = TeraDevice.get_device_by_name('Robot A')

            dev_participant = TeraDeviceParticipant()
            dev_participant.device_participant_device = device1
            dev_participant.device_participant_participant = participant1
            db.session.add(dev_participant)

            dev_participant = TeraDeviceParticipant()
            dev_participant.device_participant_device = device1
            dev_participant.device_participant_participant = participant2
            db.session.add(dev_participant)

            dev_participant = TeraDeviceParticipant()
            dev_participant.device_participant_device = device2
            dev_participant.device_participant_participant = participant2
            db.session.add(dev_participant)

            db.session.commit()

    @staticmethod
    def get_device_participant_by_id(device_participant_id: int):
        return TeraDeviceParticipant.query.filter_by(id_device_participant=device_participant_id).first()

    @staticmethod
    def query_devices_for_participant(participant_id: int):
        return TeraDeviceParticipant.query.filter_by(id_participant=participant_id).all()

    @staticmethod
    def query_participants_for_device(device_id: int):
        return TeraDeviceParticipant.query.filter_by(id_device=device_id).all()

    @staticmethod
    def query_device_participant_for_participant_device(device_id: int, participant_id: int):
        return TeraDeviceParticipant.query.filter_by(id_device=device_id, id_participant=participant_id).first()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ServerCommand.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ServerCommand.proto',
  package='opentera.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x13ServerCommand.proto\x12\x11opentera.protobuf\x1a\x19google/protobuf/any.proto\"\xa8\x01\n\rServerCommand\x12:\n\x04type\x18\x01 \x01(\x0e\x32,.opentera.protobuf.ServerCommand.CommandType\x12\"\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\"7\n\x0b\x43ommandType\x12\x0f\n\x0b\x43MD_UNKNOWN\x10\x00\x12\x17\n\x13\x43MD_REGISTER_DEVICE\x10\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])



_SERVERCOMMAND_COMMANDTYPE = _descriptor.EnumDescriptor(
  name='CommandType',
  full_name='opentera.protobuf.ServerCommand.CommandType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CMD_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CMD_REGISTER_DEVICE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=183,
  serialized_end=238,
)
_sym_db.RegisterEnumDescriptor(_SERVERCOMMAND_COMMANDTYPE)


_SERVERCOMMAND = _descriptor.Descriptor(
  name='ServerCommand',
  full_name='opentera.protobuf.ServerCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='opentera.protobuf.ServerCommand.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='opentera.protobuf.ServerCommand.data', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SERVERCOMMAND_COMMANDTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=70,
  serialized_end=238,
)

_SERVERCOMMAND.fields_by_name['type'].enum_type = _SERVERCOMMAND_COMMANDTYPE
_SERVERCOMMAND.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_SERVERCOMMAND_COMMANDTYPE.containing_type = _SERVERCOMMAND
DESCRIPTOR.message_types_by_name['ServerCommand'] = _SERVERCOMMAND
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ServerCommand = _reflection.GeneratedProtocolMessageType('ServerCommand', (_message.Message,), {
  'DESCRIPTOR' : _SERVERCOMMAND,
  '__module__' : 'ServerCommand_pb2'
  # @@protoc_insertion_point(class_scope:opentera.protobuf.ServerCommand)
  })
_sym_db.RegisterMessage(ServerCommand)


# @@protoc_insertion_point(module_scope)



from flask_sqlalchemy import SQLAlchemy, event
import inspect
import datetime
# import uuid
import time
import sqlalchemy.sql.sqltypes

db = SQLAlchemy()


class BaseModel:

    version_id = db.Column(db.BigInteger, nullable=False, default=time.time()*1000)

    # Using timestamp as version tracker - multiplying by 1000 to keep ms part without using floats (which seems to
    # cause problems with the mapper)
    __mapper_args__ = {
        'version_id_col': version_id,
        'version_id_generator': lambda version: int(time.time()*1000)  # uuid.uuid4().hex
    }

    def to_json(self, ignore_fields=None):
        if ignore_fields is None:
            ignore_fields = []
        pr = {}
        for name in dir(self):
            if self.is_valid_property_name(name) and name not in ignore_fields:
                value = getattr(self, name)
                if self.is_valid_property_value(value):
                    if isinstance(value, datetime.datetime):
                        value = value.isoformat()
                    pr[name] = value
        return pr

    def from_json(self, json, ignore_fields=None):
        if ignore_fields is None:
            ignore_fields = []
        for name in json:
            if name not in ignore_fields:
                if hasattr(self, name):
                    # Test for datetime as string
                    # This is a fix for SQLITE that does not convert str to datetime automatically
                    if isinstance(self.__table__.columns[name].type, sqlalchemy.sql.sqltypes.TIMESTAMP) \
                            and isinstance(json[name], str) and len(json[name]) > 0:
                        setattr(self, name, datetime.datetime.fromisoformat(json[name]))
                    else:
                        setattr(self, name, json[name])
                else:
                    print('Attribute ' + name + ' not found.')

    def to_json_create_event(self):
        # Default is None, will not be sent
        return None

    def to_json_update_event(self):
        # Default is None, will not be sent
        return None

    def to_json_delete_event(self):
        # Default is None, will not be sent
        return None

    @staticmethod
    def is_valid_property_name(name: str) -> bool:
        return not name.startswith('__') and not name.startswith('_') and not name.startswith('query') and \
               not name.startswith('metadata') and name != 'version_id'

    @staticmethod
    def is_valid_property_value(value: str) -> bool:
        return not inspect.ismethod(value) and not inspect.isfunction(value)

    @classmethod
    def clean_values(cls, values: dict):
        # This method is used to remove item from the values dict that are not properties of the object
        obj_properties = list()

        # Build available properties
        for name in dir(cls):
            value = getattr(cls, name)
            if cls.is_valid_property_name(name) and cls.is_valid_property_value(value):
                obj_properties.append(name)

        # Remove any property not in the available list
        clean_values = values.copy()
        for value in values:
            if value not in obj_properties:
                del clean_values[value]

        return clean_values

    @classmethod
    def get_count(cls):
        count = db.session.query(cls).count()
        return count

    @classmethod
    def get_primary_key_name(cls) -> str:
        from sqlalchemy import inspect
        return inspect(cls).primary_key[0].name

    @classmethod
    def get_model_name(cls) -> str:
        key = cls.get_primary_key_name()
        key_split = key.split('_', 1)
        return key_split[-1]

    @classmethod
    def update(cls, update_id: int, values: dict):
        values = cls.clean_values(values)
        update_obj = cls.query.filter(getattr(cls, cls.get_primary_key_name()) == update_id).first()  # .update(values)
        update_obj.from_json(values)
        cls.commit()

    @classmethod
    def commit(cls):
        db.session.commit()

    @classmethod
    def insert(cls, db_object):
        # Clear primary key value
        setattr(db_object, cls.get_primary_key_name(), None)

        # Add to database session and commit
        db.session.add(db_object)
        cls.commit()

    @classmethod
    def delete(cls, id_todel):
        delete_obj = cls.query.filter(getattr(cls, cls.get_primary_key_name()) == id_todel).first()
        if delete_obj:
            db.session.delete(delete_obj)
            cls.commit()

    @classmethod
    def query_with_filters(cls, filters=None):
        if filters is None:
            filters = dict()

        return cls.query.filter_by(**filters).all()

        # query = cls.query
        #
        # # Direct elements from that model
        # for filter_item in filters:
        #     if filter_item in dir(cls) and filter_item is not None:
        #         # Filter item is a direct property
        #         query = query.filter_by(**filter_item)
        #     else:
        #         # Check for joined models
        #         pass
        #
        # return query.all()

    @classmethod
    def count_with_filters(cls, filters=None):
        if filters is None:
            filters = dict()

        return cls.query.filter_by(**filters).count()

    @classmethod
    def get_json_schema(cls) -> dict:
        # Get model prefix (name)
        model_name = cls.get_model_name()

        # Browse each
        pr_dict = dict()
        for name in dir(cls):
            value = getattr(cls, name)
            if cls.is_valid_property_name(name) and cls.is_valid_property_value(value) and \
                    (name.startswith(model_name) or name.startswith('id')):
                # Ok so far, do we have a column and not a relationship or something else?
                if 'ColumnProperty' in str(type(value.prop)):
                    # Get correct data type
                    data_type = 'object'
                    data_format = None
                    column_type = str(value.prop.columns[0].type).lower()
                    default_value = value.prop.columns[0].default
                    if 'string' in column_type or 'timestamp' in column_type or 'varchar' in column_type:
                        data_type = 'string'
                        if 'uuid' in name:
                            data_format = 'uuid'
                        if 'timestamp' in column_type:
                            data_format = 'date-time'
                    if 'integer' in column_type:
                        data_type = 'integer'
                    if 'boolean' in column_type:
                        data_type = 'boolean'

                    pr_dict[name] = {'type': data_type, 'required': not value.prop.columns[0].nullable}
                    if data_format:
                        pr_dict[name]['format'] = data_format
                    if default_value:
                        if hasattr(default_value, 'arg'):
                            pr_dict[name]['default'] = default_value.arg

        schema = {model_name: {'properties': pr_dict, 'type': 'object'}}

        return schema

    @classmethod
    def validate_required_fields(cls, json_data: dict, ignore_fields: list = None):
        if not ignore_fields:
            ignore_fields = []

        # Get model prefix (name)
        model_name = cls.get_model_name()
        missing_fields = []

        # Browse each
        pr_dict = dict()
        for name in dir(cls):
            value = getattr(cls, name)
            if cls.is_valid_property_name(name) and cls.is_valid_property_value(value) and \
                    (name.startswith(model_name) or name.startswith('id')):
                # Ok so far, check if column is required or not
                if 'ColumnProperty' in str(type(value.prop)):
                    if not value.prop.columns[0].nullable and name not in json_data and name not in ignore_fields:
                        missing_fields.append(name)

        return missing_fields



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: UserEvent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='UserEvent.proto',
  package='opentera.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0fUserEvent.proto\x12\x11opentera.protobuf\"\x9a\x02\n\tUserEvent\x12\x11\n\tuser_uuid\x18\x01 \x01(\t\x12\x34\n\x04type\x18\x02 \x01(\x0e\x32&.opentera.protobuf.UserEvent.EventType\x12\x15\n\ruser_fullname\x18\x03 \x01(\t\"\xac\x01\n\tEventType\x12\x10\n\x0cUSER_UNKNOWN\x10\x00\x12\x12\n\x0eUSER_CONNECTED\x10\x01\x12\x15\n\x11USER_DISCONNECTED\x10\x02\x12\x10\n\x0cUSER_DELETED\x10\x03\x12\x0e\n\nUSER_ADDED\x10\x04\x12\x10\n\x0cUSER_UPDATED\x10\x05\x12\x17\n\x13USER_JOINED_SESSION\x10\x06\x12\x15\n\x11USER_LEFT_SESSION\x10\x07\x62\x06proto3'
)



_USEREVENT_EVENTTYPE = _descriptor.EnumDescriptor(
  name='EventType',
  full_name='opentera.protobuf.UserEvent.EventType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='USER_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_CONNECTED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_DISCONNECTED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_DELETED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_ADDED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_UPDATED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_JOINED_SESSION', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER_LEFT_SESSION', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=149,
  serialized_end=321,
)
_sym_db.RegisterEnumDescriptor(_USEREVENT_EVENTTYPE)


_USEREVENT = _descriptor.Descriptor(
  name='UserEvent',
  full_name='opentera.protobuf.UserEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_uuid', full_name='opentera.protobuf.UserEvent.user_uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='opentera.protobuf.UserEvent.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_fullname', full_name='opentera.protobuf.UserEvent.user_fullname', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _USEREVENT_EVENTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39,
  serialized_end=321,
)

_USEREVENT.fields_by_name['type'].enum_type = _USEREVENT_EVENTTYPE
_USEREVENT_EVENTTYPE.containing_type = _USEREVENT
DESCRIPTOR.message_types_by_name['UserEvent'] = _USEREVENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UserEvent = _reflection.GeneratedProtocolMessageType('UserEvent', (_message.Message,), {
  'DESCRIPTOR' : _USEREVENT,
  '__module__' : 'UserEvent_pb2'
  # @@protoc_insertion_point(class_scope:opentera.protobuf.UserEvent)
  })
_sym_db.RegisterMessage(UserEvent)


# @@protoc_insertion_point(module_scope)



from opentera.db.Base import db, BaseModel


class TeraDeviceProject(db.Model, BaseModel):
    __tablename__ = 't_devices_projects'
    id_device_project = db.Column(db.Integer, db.Sequence('id_device_project_sequence'), primary_key=True,
                                  autoincrement=True)
    id_device = db.Column(db.Integer, db.ForeignKey("t_devices.id_device", ondelete='cascade'), nullable=False)
    id_project = db.Column(db.Integer, db.ForeignKey("t_projects.id_project", ondelete='cascade'), nullable=False)

    device_project_project = db.relationship("TeraProject")
    device_project_device = db.relationship("TeraDevice")

    def to_json(self, ignore_fields=[], minimal=False):
        ignore_fields.extend(['device_project_project', 'device_project_device'])

        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraDevice import TeraDevice
            from opentera.db.models.TeraProject import TeraProject
            device1 = TeraDevice.get_device_by_name('Apple Watch #W05P1')
            device2 = TeraDevice.get_device_by_name('Kit Télé #1')
            device3 = TeraDevice.get_device_by_name('Robot A')
            project1 = TeraProject.get_project_by_projectname('Default Project #1')
            project2 = TeraProject.get_project_by_projectname('Default Project #2')
            project3 = TeraProject.get_project_by_projectname('Secret Project #1')

            dev_proj = TeraDeviceProject()
            dev_proj.device_project_device = device1
            dev_proj.device_project_project = project1
            db.session.add(dev_proj)

            dev_proj = TeraDeviceProject()
            dev_proj.device_project_device = device2
            dev_proj.device_project_project = project1
            db.session.add(dev_proj)

            dev_proj = TeraDeviceProject()
            dev_proj.device_project_device = device1
            dev_proj.device_project_project = project3
            db.session.add(dev_proj)

            db.session.commit()

    @staticmethod
    def get_device_project_by_id(device_project_id: int):
        return TeraDeviceProject.query.filter_by(id_device_project=device_project_id).first()

    @staticmethod
    def get_device_project_id_for_device_and_project(device_id: int, project_id: int):
        return TeraDeviceProject.query.filter_by(id_project=project_id, id_device=device_id).first()

    @staticmethod
    def query_devices_for_project(project_id: int):
        return TeraDeviceProject.query.filter_by(id_project=project_id).all()

    @staticmethod
    def query_projects_for_device(device_id: int):
        return TeraDeviceProject.query.filter_by(id_device=device_id).all()

    @staticmethod
    def query_sites_for_device(device_id: int) -> list:
        from opentera.db.models.TeraProject import TeraProject
        return TeraDeviceProject.query.filter_by(id_device=device_id).join(TeraDeviceProject.device_project_project)\
            .join(TeraProject.project_site).all()
        # TeraSite.query.join(TeraSite.site_projects).join(TeraDeviceProject.device_project_project)\
        # .filter(id_device=device_id).all()

    @staticmethod
    def query_devices_for_site(site_id: int) -> list:
        from opentera.db.models.TeraDevice import TeraDevice
        return TeraDeviceProject.query.join(TeraDeviceProject.device_project_project).filter_by(id_site=site_id).all()
        # return TeraDevice.query.join(TeraDevice.device_projects).filter_by(id_site=site_id).all()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Result.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='Result.proto',
  package='opentera.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0cResult.proto\x12\x11opentera.protobuf\"\xb2\x01\n\x06Result\x12\x32\n\x04type\x18\x01 \x01(\x0e\x32$.opentera.protobuf.Result.ResultType\x12\x0c\n\x04\x63ode\x18\x02 \x01(\x03\x12\x0f\n\x07message\x18\x03 \x01(\t\"U\n\nResultType\x12\x12\n\x0eRESULT_UNKNOWN\x10\x00\x12\r\n\tRESULT_OK\x10\x01\x12\x10\n\x0cRESULT_ERROR\x10\x02\x12\x12\n\x0eRESULT_WARNING\x10\x03\x62\x06proto3'
)



_RESULT_RESULTTYPE = _descriptor.EnumDescriptor(
  name='ResultType',
  full_name='opentera.protobuf.Result.ResultType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RESULT_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RESULT_OK', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RESULT_ERROR', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RESULT_WARNING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=129,
  serialized_end=214,
)
_sym_db.RegisterEnumDescriptor(_RESULT_RESULTTYPE)


_RESULT = _descriptor.Descriptor(
  name='Result',
  full_name='opentera.protobuf.Result',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='opentera.protobuf.Result.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='opentera.protobuf.Result.code', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='opentera.protobuf.Result.message', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RESULT_RESULTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=36,
  serialized_end=214,
)

_RESULT.fields_by_name['type'].enum_type = _RESULT_RESULTTYPE
_RESULT_RESULTTYPE.containing_type = _RESULT
DESCRIPTOR.message_types_by_name['Result'] = _RESULT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Result = _reflection.GeneratedProtocolMessageType('Result', (_message.Message,), {
  'DESCRIPTOR' : _RESULT,
  '__module__' : 'Result_pb2'
  # @@protoc_insertion_point(class_scope:opentera.protobuf.Result)
  })
_sym_db.RegisterMessage(Result)


# @@protoc_insertion_point(module_scope)



from opentera.db.Base import db, BaseModel
from enum import Enum, unique
from flask_babel import gettext


class TeraSessionType(db.Model, BaseModel):
    @unique
    class SessionCategoryEnum(Enum):
        SERVICE = 1
        DATACOLLECT = 2
        FILETRANSFER = 3
        PROTOCOL = 4

        def describe(self):
            return self.name, self.value

    __tablename__ = 't_sessions_types'
    id_session_type = db.Column(db.Integer, db.Sequence('id_session_type_sequence'), primary_key=True,
                                autoincrement=True)
    id_service = db.Column(db.Integer, db.ForeignKey('t_services.id_service', ondelete='cascade'), nullable=True)
    session_type_name = db.Column(db.String, nullable=False, unique=False)
    session_type_online = db.Column(db.Boolean, nullable=False)
    session_type_config = db.Column(db.String, nullable=True)
    session_type_color = db.Column(db.String(7), nullable=False)
    session_type_category = db.Column(db.Integer, nullable=False)

    session_type_session_type_projects = db.relationship("TeraSessionTypeProject", passive_deletes=True)
    session_type_projects = db.relationship("TeraProject", secondary="t_sessions_types_projects",
                                            back_populates="project_session_types")

    session_type_service = db.relationship("TeraService")

    session_type_sessions = db.relationship("TeraSession", passive_deletes=True)

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.extend(['session_type_projects', 'session_type_devices_types', 'SessionCategoryEnum',
                              'session_type_service', 'session_type_sessions', 'session_type_session_type_projects'])
        if minimal:
            ignore_fields.extend(['session_type_online',
                                  'session_type_profile',
                                  'session_type_config'])
        rval = super().to_json(ignore_fields=ignore_fields)

        if not minimal:
            # Also includes service key and uuid
            if self.session_type_service:
                rval['session_type_service_key'] = self.session_type_service.service_key
                rval['session_type_service_uuid'] = self.session_type_service.service_uuid
        return rval

    @staticmethod
    def create_defaults(test=False):
        # from opentera.db.models.TeraProject import TeraProject
        # from opentera.db.models.TeraDeviceType import TeraDeviceType
        from opentera.db.models.TeraService import TeraService

        # type_project = TeraProject.get_project_by_projectname('Default Project #1')
        video_session = TeraSessionType()
        video_session.session_type_name = "Suivi vidéo"
        video_session.session_type_online = True
        video_session.session_type_config = ""
        video_session.session_type_color = "#00FF00"
        video_session.session_type_category = TeraSessionType.SessionCategoryEnum.SERVICE.value
        video_session.id_service = TeraService.get_service_by_key('VideoRehabService').id_service
        # video_session.session_type_projects = [type_project]
        # video_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.VIDEOCONFERENCE.value))]
        db.session.add(video_session)

        sensor_session = TeraSessionType()
        sensor_session.session_type_name = "Données Capteur"
        sensor_session.session_type_online = False
        sensor_session.session_type_config = ""
        sensor_session.session_type_color = "#0000FF"
        sensor_session.session_type_category = TeraSessionType.SessionCategoryEnum.FILETRANSFER.value
        # sensor_session.session_type_projects = [type_project]
        # sensor_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.SENSOR.value))]
        db.session.add(sensor_session)

        vsensor_session = TeraSessionType()
        vsensor_session.session_type_name = "Collecte données"
        vsensor_session.session_type_online = True
        vsensor_session.session_type_config = ""
        vsensor_session.session_type_color = "#00FFFF"
        # vsensor_session.session_type_projects = [type_project]
        vsensor_session.session_type_category = TeraSessionType.SessionCategoryEnum.DATACOLLECT.value
        # vsensor_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.SENSOR.value)), TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.VIDEOCONFERENCE.value))]
        db.session.add(vsensor_session)

        robot_session = TeraSessionType()
        robot_session.session_type_name = "Exercices individuels"
        robot_session.session_type_online = False
        robot_session.session_type_config = ""
        robot_session.session_type_color = "#FF00FF"
        # robot_session.session_type_projects = [type_project]
        robot_session.session_type_category = TeraSessionType.SessionCategoryEnum.PROTOCOL.value
        # robot_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.ROBOT.value))]
        db.session.add(robot_session)

        bureau_session = TeraSessionType()
        bureau_session.session_type_name = "Bureau Actif"
        bureau_session.session_type_online = False
        bureau_session.session_type_config = ""
        bureau_session.session_type_color = "#FF00FF"
        bureau_session.session_type_category = TeraSessionType.SessionCategoryEnum.SERVICE.value
        # robot_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.ROBOT.value))]
        bureau_session.id_service = TeraService.get_service_by_key('BureauActif').id_service
        db.session.add(bureau_session)

        db.session.commit()

    @staticmethod
    def get_session_type_by_id(ses_type_id: int):
        return TeraSessionType.query.filter_by(id_session_type=ses_type_id).first()

    @staticmethod
    def get_session_types_for_service(id_service: int):
        return TeraSessionType.query.filter_by(id_service=id_service).all()

    @staticmethod
    def get_category_name(category: SessionCategoryEnum):
        name = gettext('Unknown')
        if category == TeraSessionType.SessionCategoryEnum.SERVICE:
            name = gettext('Service')
        if category == TeraSessionType.SessionCategoryEnum.FILETRANSFER:
            name = gettext('File Transfer')
        if category == TeraSessionType.SessionCategoryEnum.DATACOLLECT:
            name = gettext('Data Collect')
        if category == TeraSessionType.SessionCategoryEnum.PROTOCOL:
            name = gettext('Protocol')

        return name

    @classmethod
    def update(cls, id_st: int, values: dict):
        # Set service id to None if setted to 0
        if 'id_service' in values:
            if values['id_service'] == 0:
                del values['id_service']

        super().update(id_st, values)

    @classmethod
    def insert(cls, st):
        if st.id_service == 0:
            st.id_service = None

        super().insert(st)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: JoinSessionReplyEvent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='JoinSessionReplyEvent.proto',
  package='opentera.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1bJoinSessionReplyEvent.proto\x12\x11opentera.protobuf\"\xb3\x02\n\x15JoinSessionReplyEvent\x12\x14\n\x0csession_uuid\x18\x01 \x01(\t\x12\x11\n\tuser_uuid\x18\x02 \x01(\t\x12\x18\n\x10participant_uuid\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65vice_uuid\x18\x04 \x01(\t\x12\x46\n\njoin_reply\x18\x05 \x01(\x0e\x32\x32.opentera.protobuf.JoinSessionReplyEvent.ReplyType\x12\x11\n\treply_msg\x18\x06 \x01(\t\"g\n\tReplyType\x12\x11\n\rREPLY_UNKNOWN\x10\x00\x12\x12\n\x0eREPLY_ACCEPTED\x10\x01\x12\x10\n\x0cREPLY_DENIED\x10\x02\x12\x0e\n\nREPLY_BUSY\x10\x03\x12\x11\n\rREPLY_TIMEOUT\x10\x04\x62\x06proto3'
)



_JOINSESSIONREPLYEVENT_REPLYTYPE = _descriptor.EnumDescriptor(
  name='ReplyType',
  full_name='opentera.protobuf.JoinSessionReplyEvent.ReplyType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REPLY_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPLY_ACCEPTED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPLY_DENIED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPLY_BUSY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPLY_TIMEOUT', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=255,
  serialized_end=358,
)
_sym_db.RegisterEnumDescriptor(_JOINSESSIONREPLYEVENT_REPLYTYPE)


_JOINSESSIONREPLYEVENT = _descriptor.Descriptor(
  name='JoinSessionReplyEvent',
  full_name='opentera.protobuf.JoinSessionReplyEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_uuid', full_name='opentera.protobuf.JoinSessionReplyEvent.session_uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_uuid', full_name='opentera.protobuf.JoinSessionReplyEvent.user_uuid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='participant_uuid', full_name='opentera.protobuf.JoinSessionReplyEvent.participant_uuid', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='device_uuid', full_name='opentera.protobuf.JoinSessionReplyEvent.device_uuid', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='join_reply', full_name='opentera.protobuf.JoinSessionReplyEvent.join_reply', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reply_msg', full_name='opentera.protobuf.JoinSessionReplyEvent.reply_msg', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _JOINSESSIONREPLYEVENT_REPLYTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=51,
  serialized_end=358,
)

_JOINSESSIONREPLYEVENT.fields_by_name['join_reply'].enum_type = _JOINSESSIONREPLYEVENT_REPLYTYPE
_JOINSESSIONREPLYEVENT_REPLYTYPE.containing_type = _JOINSESSIONREPLYEVENT
DESCRIPTOR.message_types_by_name['JoinSessionReplyEvent'] = _JOINSESSIONREPLYEVENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

JoinSessionReplyEvent = _reflection.GeneratedProtocolMessageType('JoinSessionReplyEvent', (_message.Message,), {
  'DESCRIPTOR' : _JOINSESSIONREPLYEVENT,
  '__module__' : 'JoinSessionReplyEvent_pb2'
  # @@protoc_insertion_point(class_scope:opentera.protobuf.JoinSessionReplyEvent)
  })
_sym_db.RegisterMessage(JoinSessionReplyEvent)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: RPCMessage.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='RPCMessage.proto',
  package='opentera.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x10RPCMessage.proto\x12\x11opentera.protobuf\"\xb3\x01\n\x05Value\x12\x16\n\x0c\x64ouble_value\x18\x01 \x01(\x01H\x00\x12\x15\n\x0b\x66loat_value\x18\x02 \x01(\x02H\x00\x12\x13\n\tint_value\x18\x03 \x01(\x03H\x00\x12\x14\n\nuint_value\x18\x04 \x01(\x04H\x00\x12\x14\n\nbool_value\x18\x05 \x01(\x08H\x00\x12\x15\n\x0b\x62ytes_value\x18\x06 \x01(\x0cH\x00\x12\x16\n\x0cstring_value\x18\x07 \x01(\tH\x00\x42\x0b\n\targ_value\"u\n\nRPCMessage\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x10\n\x08reply_to\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\x04\x12\x11\n\ttimestamp\x18\x04 \x01(\x01\x12&\n\x04\x61rgs\x18\x05 \x03(\x0b\x32\x18.opentera.protobuf.Valueb\x06proto3'
)




_VALUE = _descriptor.Descriptor(
  name='Value',
  full_name='opentera.protobuf.Value',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='double_value', full_name='opentera.protobuf.Value.double_value', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='float_value', full_name='opentera.protobuf.Value.float_value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='int_value', full_name='opentera.protobuf.Value.int_value', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uint_value', full_name='opentera.protobuf.Value.uint_value', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bool_value', full_name='opentera.protobuf.Value.bool_value', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bytes_value', full_name='opentera.protobuf.Value.bytes_value', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='string_value', full_name='opentera.protobuf.Value.string_value', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='arg_value', full_name='opentera.protobuf.Value.arg_value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=40,
  serialized_end=219,
)


_RPCMESSAGE = _descriptor.Descriptor(
  name='RPCMessage',
  full_name='opentera.protobuf.RPCMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='method', full_name='opentera.protobuf.RPCMessage.method', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reply_to', full_name='opentera.protobuf.RPCMessage.reply_to', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='opentera.protobuf.RPCMessage.id', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='opentera.protobuf.RPCMessage.timestamp', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='args', full_name='opentera.protobuf.RPCMessage.args', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=221,
  serialized_end=338,
)

_VALUE.oneofs_by_name['arg_value'].fields.append(
  _VALUE.fields_by_name['double_value'])
_VALUE.fields_by_name['double_value'].containing_oneof = _VALUE.oneofs_by_name['arg_value']
_VALUE.oneofs_by_name['arg_value'].fields.append(
  _VALUE.fields_by_name['float_value'])
_VALUE.fields_by_name['float_value'].containing_oneof = _VALUE.oneofs_by_name['arg_value']
_VALUE.oneofs_by_name['arg_value'].fields.append(
  _VALUE.fields_by_name['int_value'])
_VALUE.fields_by_name['int_value'].containing_oneof = _VALUE.oneofs_by_name['arg_value']
_VALUE.oneofs_by_name['arg_value'].fields.append(
  _VALUE.fields_by_name['uint_value'])
_VALUE.fields_by_name['uint_value'].containing_oneof = _VALUE.oneofs_by_name['arg_value']
_VALUE.oneofs_by_name['arg_value'].fields.append(
  _VALUE.fields_by_name['bool_value'])
_VALUE.fields_by_name['bool_value'].containing_oneof = _VALUE.oneofs_by_name['arg_value']
_VALUE.oneofs_by_name['arg_value'].fields.append(
  _VALUE.fields_by_name['bytes_value'])
_VALUE.fields_by_name['bytes_value'].containing_oneof = _VALUE.oneofs_by_name['arg_value']
_VALUE.oneofs_by_name['arg_value'].fields.append(
  _VALUE.fields_by_name['string_value'])
_VALUE.fields_by_name['string_value'].containing_oneof = _VALUE.oneofs_by_name['arg_value']
_RPCMESSAGE.fields_by_name['args'].message_type = _VALUE
DESCRIPTOR.message_types_by_name['Value'] = _VALUE
DESCRIPTOR.message_types_by_name['RPCMessage'] = _RPCMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), {
  'DESCRIPTOR' : _VALUE,
  '__module__' : 'RPCMessage_pb2'
  # @@protoc_insertion_point(class_scope:opentera.protobuf.Value)
  })
_sym_db.RegisterMessage(Value)

RPCMessage = _reflection.GeneratedProtocolMessageType('RPCMessage', (_message.Message,), {
  'DESCRIPTOR' : _RPCMESSAGE,
  '__module__' : 'RPCMessage_pb2'
  # @@protoc_insertion_point(class_scope:opentera.protobuf.RPCMessage)
  })
_sym_db.RegisterMessage(RPCMessage)


# @@protoc_insertion_point(module_scope)



from opentera.db.Base import db, BaseModel
from opentera.db.models.TeraProject import TeraProject


class TeraParticipantGroup(db.Model, BaseModel):
    __tablename__ = 't_participants_groups'
    id_participant_group = db.Column(db.Integer, db.Sequence('id_participantgroup_sequence'), primary_key=True,
                                     autoincrement=True)
    id_project = db.Column(db.Integer, db.ForeignKey('t_projects.id_project', ondelete='cascade'), nullable=False)
    participant_group_name = db.Column(db.String, nullable=False, unique=False)

    participant_group_project = db.relationship('TeraProject')
    participant_group_participants = db.relationship("TeraParticipant", passive_deletes=True)

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.extend(['participant_group_project', 'participant_group_participants'])
        rval = super().to_json(ignore_fields=ignore_fields)

        if not minimal:
            rval['project_name'] = self.participant_group_project.project_name
        # rval['id_site'] = self.participant_group_project.id_site
        return rval

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_participant_group': self.id_participant_group}

    @staticmethod
    def get_participant_group_by_group_name(name: str):
        return TeraParticipantGroup.query.filter_by(participant_group_name=name).first()

    @staticmethod
    def get_participant_group_by_id(group_id: int):
        return TeraParticipantGroup.query.filter_by(id_participant_group=group_id).first()

    @staticmethod
    def get_participant_group_for_project(project_id: int):
        return TeraParticipantGroup.query.filter_by(id_project=project_id).all()

    @staticmethod
    def create_defaults(test=False):
        if test:
            base_pgroup = TeraParticipantGroup()
            base_pgroup.participant_group_name = 'Default Participant Group A'
            base_pgroup.id_project = TeraProject.get_project_by_projectname('Default Project #1').id_project
            db.session.add(base_pgroup)

            base_pgroup2 = TeraParticipantGroup()
            base_pgroup2.participant_group_name = 'Default Participant Group B'
            base_pgroup2.id_project = TeraProject.get_project_by_projectname('Default Project #2').id_project
            db.session.add(base_pgroup2)
            db.session.commit()

    @classmethod
    def delete(cls, id_todel):
        super().delete(id_todel)

        # from opentera.db.models.TeraSession import TeraSession
        # TeraSession.delete_orphaned_sessions()



from opentera.forms.TeraForm import *
from flask_babel import gettext
from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from opentera.db.models.TeraSessionType import TeraSessionType


class TeraSessionTypeConfigForm:

    @staticmethod
    def get_session_type_config_form(user_access: DBManagerTeraUserAccess):
        form = TeraForm("session_type_config")

        # Sections
        section = TeraFormSection("general", gettext("General configuration"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("session_recordable", gettext("Allow session recording"), "boolean", False,
                                      item_default=False))

        return form.to_dict()





from werkzeug.local import LocalProxy
from functools import wraps
from flask import _request_ctx_stack, request
from flask_restx import reqparse

from enum import Enum

from opentera.services.TeraUserClient import TeraUserClient
from opentera.services.TeraDeviceClient import TeraDeviceClient
from opentera.services.TeraParticipantClient import TeraParticipantClient
from opentera.services.TeraServiceClient import TeraServiceClient

# Current client identity, stacked
current_user_client = LocalProxy(lambda: getattr(_request_ctx_stack.top, 'current_user_client', None))
current_device_client = LocalProxy(lambda: getattr(_request_ctx_stack.top, 'current_device_client', None))
current_participant_client = LocalProxy(lambda: getattr(_request_ctx_stack.top, 'current_participant_client', None))
current_service_client = LocalProxy(lambda: getattr(_request_ctx_stack.top, 'current_service_client', None))


current_login_type = LocalProxy(lambda: getattr(_request_ctx_stack.top, 'current_login_type', LoginType.UNKNOWN_LOGIN))


class LoginType(Enum):
    UNKNOWN_LOGIN = 0,
    USER_LOGIN = 1,
    DEVICE_LOGIN = 2,
    PARTICIPANT_LOGIN = 3,
    SERVICE_LOGIN = 4


class ServiceAccessManager:

    api_user_token_key = None
    api_participant_token_key = None
    api_participant_static_token_key = None
    api_device_token_key = None
    api_device_static_token_key = None
    api_service_token_key = None
    token_cookie_name = 'OpenTera'
    config_man = None

    @staticmethod
    def token_required(allow_dynamic_tokens=True, allow_static_tokens=False):
        def wrap(f):
            def decorated(*args, **kwargs):
                # We support 3 authentication scheme: token in url, cookie and authorization header
                token = None
                ######################
                # AUTHORIZATION HEADER
                if 'Authorization' in request.headers:
                    try:
                        # Default whitespace as separator, 1 split max
                        scheme, atoken = request.headers['Authorization'].split(None, 1)
                    except ValueError:
                        # malformed Authorization header
                        return 'Forbidden', 403

                    # Verify scheme and token
                    if scheme == 'OpenTera':
                        token = atoken
                if token is None:
                    #################
                    # TOKEN PARAMETER ?
                    parser = reqparse.RequestParser()
                    parser.add_argument('token', type=str, help='Device, participant or user token', required=False)

                    # Parse arguments
                    request_args = parser.parse_args(strict=False)

                    # Verify token in params
                    if 'token' in request_args:
                        token = request_args['token']
                if token is None:
                    ###############
                    # COOKIE TOKEN
                    if ServiceAccessManager.token_cookie_name in request.cookies:
                        token = request.cookies[ServiceAccessManager.token_cookie_name]

                #########################
                # Verify token from redis
                import jwt

                # USER TOKEN MANAGEMENT
                if allow_dynamic_tokens:  # Users only have dynamic tokens, if the flag isn't set, don't validate token!
                    try:
                        token_dict = jwt.decode(token, ServiceAccessManager.api_user_token_key, algorithms='HS256')
                    except jwt.PyJWTError as e:
                        # Not a user, or invalid token, will continue...
                        pass
                    else:
                        # User token
                        _request_ctx_stack.top.current_user_client = \
                            TeraUserClient(token_dict, token, ServiceAccessManager.config_man)
                        _request_ctx_stack.top.current_login_type = LoginType.USER_LOGIN
                        return f(*args, **kwargs)

                # DEVICE TOKEN MANAGEMENT
                if allow_dynamic_tokens:    # Check for dynamic device token
                    try:
                        token_dict = jwt.decode(token, ServiceAccessManager.api_device_token_key, algorithms='HS256')
                    except jwt.PyJWTError as e:
                        # Not a device, or invalid token, will continue...
                        pass
                    else:
                        # Device token
                        _request_ctx_stack.top.current_device_client = \
                            TeraDeviceClient(token_dict, token, ServiceAccessManager.config_man)
                        _request_ctx_stack.top.current_login_type = LoginType.DEVICE_LOGIN
                        return f(*args, **kwargs)

                if allow_static_tokens:  # Check for static device token
                    try:
                        token_dict = jwt.decode(token, ServiceAccessManager.api_device_static_token_key, algorithms='HS256')
                    except jwt.PyJWTError as e:
                        # Not a device, or invalid token, will continue...
                        pass
                    else:
                        # Device token
                        _request_ctx_stack.top.current_device_client = \
                            TeraDeviceClient(token_dict, token, ServiceAccessManager.config_man)
                        _request_ctx_stack.top.current_login_type = LoginType.DEVICE_LOGIN
                        return f(*args, **kwargs)

                # PARTICIPANT TOKEN MANAGEMENT
                if allow_dynamic_tokens:  # Check for dynamic participant token
                    try:
                        token_dict = jwt.decode(token, ServiceAccessManager.api_participant_token_key, algorithms='HS256')
                    except jwt.PyJWTError as e:
                        # Not a participant, or invalid token, will continue...
                        pass
                    else:
                        # Participant token
                        _request_ctx_stack.top.current_participant_client = \
                            TeraParticipantClient(token_dict, token, ServiceAccessManager.config_man)
                        _request_ctx_stack.top.current_login_type = LoginType.PARTICIPANT_LOGIN
                        return f(*args, **kwargs)

                if allow_static_tokens:  # Check for static participant token
                    try:
                        token_dict = jwt.decode(token, ServiceAccessManager.api_participant_static_token_key,
                                                algorithms='HS256')
                    except jwt.PyJWTError as e:
                        # Not a participant, or invalid token, will continue...
                        pass
                    else:
                        # Participant token
                        _request_ctx_stack.top.current_participant_client = \
                            TeraParticipantClient(token_dict, token, ServiceAccessManager.config_man)
                        _request_ctx_stack.top.current_login_type = LoginType.PARTICIPANT_LOGIN
                        return f(*args, **kwargs)

                return 'Forbidden', 403

            return decorated
        return wrap

    # @staticmethod
    # def static_token_required(f):
    #     @wraps(f)
    #     def decorated(*args, **kwargs):
    #         # We support 3 authentication scheme: token in url, cookie and authorization header
    #         token = None
    #         ######################
    #         # AUTHORIZATION HEADER
    #         if 'Authorization' in request.headers:
    #             try:
    #                 # Default whitespace as separator, 1 split max
    #                 scheme, atoken = request.headers['Authorization'].split(None, 1)
    #             except ValueError:
    #                 # malformed Authorization header
    #                 return 'Forbidden', 403
    #
    #             # Verify scheme and token
    #             if scheme == 'OpenTera':
    #                 token = atoken
    #         if token is None:
    #             #################
    #             # TOKEN PARAMETER ?
    #             parser = reqparse.RequestParser()
    #             parser.add_argument('token', type=str, help='Device, participant or user token', required=False)
    #
    #             # Parse arguments
    #             request_args = parser.parse_args(strict=False)
    #
    #             # Verify token in params
    #             if 'token' in request_args:
    #                 token = request_args['token']
    #
    #         #########################
    #         # Verify token from redis
    #         import jwt
    #
    #         # Do we have a device token?
    #         try:
    #             token_dict = jwt.decode(token, ServiceAccessManager.api_device_static_token_key, algorithms='HS256')
    #         except jwt.PyJWTError as e:
    #             # Not a device, or invalid token, will continue...
    #             pass
    #         else:
    #             # Device token
    #             _request_ctx_stack.top.current_device_client = \
    #                 TeraDeviceClient(token_dict, token, ServiceAccessManager.config_man)
    #             _request_ctx_stack.top.current_login_type = LoginType.DEVICE_LOGIN
    #             return f(*args, **kwargs)
    #
    #         # Do we have a participant token?
    #         try:
    #             token_dict = jwt.decode(token, ServiceAccessManager.api_participant_static_token_key,
    #                                     algorithms='HS256')
    #         except jwt.PyJWTError as e:
    #             # Not a participant, or invalid token, will continue...
    #             pass
    #         else:
    #             # Participant token
    #             _request_ctx_stack.top.current_participant_client = \
    #                 TeraParticipantClient(token_dict, token, ServiceAccessManager.config_man)
    #             _request_ctx_stack.top.current_login_type = LoginType.PARTICIPANT_LOGIN
    #             return f(*args, **kwargs)
    #
    #         return 'Forbidden', 403
    #
    #     return decorated

    @staticmethod
    def service_token_required(f):
        @wraps(f)
        def decorated(*args, **kwargs):
            # We support token auth. only.
            token = None
            ######################
            # AUTHORIZATION HEADER
            if 'Authorization' in request.headers:
                try:
                    # Default whitespace as separator, 1 split max
                    scheme, atoken = request.headers['Authorization'].split(None, 1)
                except ValueError:
                    # malformed Authorization header
                    return 'Forbidden', 403

                # Verify scheme and token
                if scheme == 'OpenTera':
                    token = atoken

            # Verify token from with service api key
            import jwt

            try:
                token_dict = jwt.decode(token, ServiceAccessManager.api_service_token_key, algorithms='HS256')
            except jwt.PyJWTError as e:
                # Not a device, or invalid token, will continue...
                pass
            else:
                # Service token
                _request_ctx_stack.top.current_service_client = \
                    TeraServiceClient(token_dict, token, ServiceAccessManager.config_man)
                _request_ctx_stack.top.current_login_type = LoginType.SERVICE_LOGIN
                return f(*args, **kwargs)

            return 'Forbidden', 403

        return decorated



from opentera.db.Base import db, BaseModel

from enum import Enum, unique
from datetime import datetime, timedelta
import random


class TeraSessionEvent(db.Model, BaseModel):
    @unique
    class SessionEventTypes(Enum):
        GENERAL_ERROR = 0
        GENERAL_INFO = 1
        GENERAL_WARNING = 2
        SESSION_START = 3
        SESSION_STOP = 4
        DEVICE_ON_CHARGE = 5
        DEVICE_OFF_CHARGE = 6
        DEVICE_LOW_BATT = 7
        DEVICE_STORAGE_LOW = 8
        DEVICE_STORAGE_FULL = 9
        DEVICE_EVENT = 10
        USER_EVENT = 11
        SESSION_JOIN = 12
        SESSION_LEAVE = 13
        SESSION_JOIN_REFUSED = 14

    __tablename__ = 't_sessions_events'
    id_session_event = db.Column(db.Integer, db.Sequence('id_session_events_sequence'), primary_key=True,
                                 autoincrement=True)
    id_session = db.Column(db.Integer, db.ForeignKey('t_sessions.id_session', ondelete='cascade'), nullable=False)
    id_session_event_type = db.Column(db.Integer, nullable=False)
    session_event_datetime = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    session_event_text = db.Column(db.String, nullable=True)
    session_event_context = db.Column(db.String, nullable=True)

    session_event_session = db.relationship('TeraSession')

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['session_event_session', 'SessionEventTypes'])
        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraSession import TeraSession

            base_session = TeraSession.get_session_by_name('Séance #1')
            for i in range(12):
                event = TeraSessionEvent()
                event.session_event_session = base_session
                event.id_session_event_type = i
                event.session_event_datetime = datetime.now() - timedelta(hours=random.randint(0, 10)) - timedelta(
                    minutes=random.randint(0, 45))
                event.session_event_context = 'Défaut'
                event.session_event_text = str(TeraSessionEvent.SessionEventTypes(i))
                db.session.add(event)
            db.session.commit()

    @staticmethod
    def get_session_event_by_id(event_id: int):
        return TeraSessionEvent.query.filter_by(id_session_event=event_id).first()

    @staticmethod
    def get_events_for_session(id_session: int):
        from .TeraSession import TeraSession
        return db.session.query(TeraSessionEvent).join(TeraSessionEvent.session_event_session)\
            .filter(TeraSession.id_session == id_session)



from opentera.db.Base import db, BaseModel
from flask_babel import gettext

import uuid


class TeraService(db.Model, BaseModel):
    __tablename__ = 't_services'

    id_service = db.Column(db.Integer, db.Sequence('id_service_sequence'), primary_key=True, autoincrement=True)
    service_uuid = db.Column(db.String(36), nullable=False, unique=True)
    service_name = db.Column(db.String, nullable=False)
    service_key = db.Column(db.String, nullable=False, unique=True)
    service_hostname = db.Column(db.String, nullable=False)
    service_port = db.Column(db.Integer, nullable=False)
    service_endpoint = db.Column(db.String, nullable=False)
    service_clientendpoint = db.Column(db.String, nullable=False)
    service_endpoint_user = db.Column(db.String, nullable=True)
    service_endpoint_participant = db.Column(db.String, nullable=True)
    service_endpoint_device = db.Column(db.String, nullable=True)
    service_enabled = db.Column(db.Boolean, nullable=False, default=False)
    service_system = db.Column(db.Boolean, nullable=False, default=False)
    service_editable_config = db.Column(db.Boolean, nullable=False, default=False)
    service_default_config = db.Column(db.String, nullable=True, default='{}')

    service_roles = db.relationship('TeraServiceRole')
    service_projects = db.relationship('TeraServiceProject')

    def __init__(self):
        pass

    def __str__(self):
        return '<TeraService ' + str(self.service_name) + ' >'

    def __repr__(self):
        return self.__str__()

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_roles', 'service_projects', 'service_system'])

        if minimal:
            ignore_fields.extend(['service_default_config'])

        json_service = super().to_json(ignore_fields=ignore_fields)
        if not minimal:
            # Add roles for that service
            roles = []
            for role in self.service_roles:
                roles.append(role.to_json())
            json_service['service_roles'] = roles
            # Add projects for that service
            service_projects = []
            for sp in self.service_projects:
                service_projects.append(sp.to_json())
            json_service['service_projects'] = service_projects

        if not self.service_endpoint_user:
            del json_service['service_endpoint_user']

        if not self.service_endpoint_participant:
            del json_service['service_endpoint_participant']

        if not self.service_endpoint_device:
            del json_service['service_endpoint_device']

        return json_service

    def get_token(self, token_key: str, expiration=3600):
        import time
        import jwt
        # Creating token with user info
        now = time.time()
        payload = {
            'iat': int(now),
            'exp': int(now) + expiration,
            'iss': 'TeraServer',
            'service_uuid': self.service_uuid
        }

        return jwt.encode(payload, token_key, algorithm='HS256')

    @staticmethod
    def get_service_by_key(key: str):
        service = TeraService.query.filter_by(service_key=key).first()

        if service:
            return service

        return None

    @staticmethod
    def get_service_by_uuid(p_uuid: str):
        service = TeraService.query.filter_by(service_uuid=p_uuid).first()

        if service:
            return service

        return None

    @staticmethod
    def get_service_by_name(name: str):
        return TeraService.query.filter_by(service_name=name).first()

    @staticmethod
    def get_service_by_id(s_id: int):
        return TeraService.query.filter_by(id_service=s_id).first()

    @staticmethod
    def get_openteraserver_service():
        return TeraService.get_service_by_key('OpenTeraServer')

    @staticmethod
    def create_defaults(test=False):
        new_service = TeraService()
        new_service.service_uuid = '00000000-0000-0000-0000-000000000001'
        new_service.service_key = 'OpenTeraServer'
        new_service.service_name = 'OpenTera Server'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4040
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/'
        new_service.service_endpoint_participant = '/participant'
        new_service.service_endpoint_user = '/user'
        new_service.service_endpoint_device = '/device'
        new_service.service_enabled = True
        new_service.service_system = True
        new_service.service_editable_config = True
        db.session.add(new_service)

        new_service = TeraService()
        new_service.service_uuid = str(uuid.uuid4())
        new_service.service_key = 'LoggingService'
        new_service.service_name = 'Logging Service'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4041
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/log'
        new_service.service_enabled = True
        new_service.service_system = True
        db.session.add(new_service)

        new_service = TeraService()
        new_service.service_uuid = str(uuid.uuid4())
        new_service.service_key = 'FileTransferService'
        new_service.service_name = 'File Transfer Service'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4042
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/file'
        new_service.service_enabled = True
        new_service.service_system = True
        db.session.add(new_service)

        new_service = TeraService()
        new_service.service_uuid = str(uuid.uuid4())
        new_service.service_key = 'BureauActif'
        new_service.service_name = 'Bureau Actif'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4050
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/bureau'
        new_service.service_enabled = True
        db.session.add(new_service)

        new_service = TeraService()
        new_service.service_uuid = str(uuid.uuid4())
        new_service.service_key = 'VideoRehabService'
        new_service.service_name = 'Télé-réadaptation vidéo'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4070
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/rehab'
        new_service.service_endpoint_participant = '/participant'
        # Not yet implemented...
        # new_service.service_endpoint_user = '/user'
        # new_service.service_endpoint_device = '/device'
        new_service.service_enabled = True
        new_service.service_editable_config = True
        db.session.add(new_service)

        new_service = TeraService()
        new_service.service_uuid = str(uuid.uuid4())
        new_service.service_key = 'RobotTeleOperationService'
        new_service.service_name = 'Robot Teleoperation Service'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4080
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/robot'
        new_service.service_enabled = True
        db.session.add(new_service)

        db.session.commit()

    @classmethod
    def insert(cls, service):
        service.service_uuid = str(uuid.uuid4())
        super().insert(service)

    @classmethod
    def update(cls, update_id: int, values: dict):
        # Prevent changes on UUID
        if 'service_uuid' in values:
            del values['service_uuid']

        super().update(update_id, values)




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TeraModuleMessage.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='TeraModuleMessage.proto',
  package='opentera.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x17TeraModuleMessage.proto\x12\x11opentera.protobuf\x1a\x19google/protobuf/any.proto\"\xc6\x01\n\x11TeraModuleMessage\x12\x39\n\x04head\x18\x01 \x01(\x0b\x32+.opentera.protobuf.TeraModuleMessage.Header\x12\"\n\x04\x64\x61ta\x18\x02 \x03(\x0b\x32\x14.google.protobuf.Any\x1aR\n\x06Header\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x0c\n\x04time\x18\x02 \x01(\x01\x12\x0b\n\x03seq\x18\x03 \x01(\r\x12\x0e\n\x06source\x18\x04 \x01(\t\x12\x0c\n\x04\x64\x65st\x18\x05 \x01(\tb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])




_TERAMODULEMESSAGE_HEADER = _descriptor.Descriptor(
  name='Header',
  full_name='opentera.protobuf.TeraModuleMessage.Header',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='opentera.protobuf.TeraModuleMessage.Header.version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time', full_name='opentera.protobuf.TeraModuleMessage.Header.time', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='seq', full_name='opentera.protobuf.TeraModuleMessage.Header.seq', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='opentera.protobuf.TeraModuleMessage.Header.source', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dest', full_name='opentera.protobuf.TeraModuleMessage.Header.dest', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=190,
  serialized_end=272,
)

_TERAMODULEMESSAGE = _descriptor.Descriptor(
  name='TeraModuleMessage',
  full_name='opentera.protobuf.TeraModuleMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='head', full_name='opentera.protobuf.TeraModuleMessage.head', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='opentera.protobuf.TeraModuleMessage.data', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TERAMODULEMESSAGE_HEADER, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=74,
  serialized_end=272,
)

_TERAMODULEMESSAGE_HEADER.containing_type = _TERAMODULEMESSAGE
_TERAMODULEMESSAGE.fields_by_name['head'].message_type = _TERAMODULEMESSAGE_HEADER
_TERAMODULEMESSAGE.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
DESCRIPTOR.message_types_by_name['TeraModuleMessage'] = _TERAMODULEMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TeraModuleMessage = _reflection.GeneratedProtocolMessageType('TeraModuleMessage', (_message.Message,), {

  'Header' : _reflection.GeneratedProtocolMessageType('Header', (_message.Message,), {
    'DESCRIPTOR' : _TERAMODULEMESSAGE_HEADER,
    '__module__' : 'TeraModuleMessage_pb2'
    # @@protoc_insertion_point(class_scope:opentera.protobuf.TeraModuleMessage.Header)
    })
  ,
  'DESCRIPTOR' : _TERAMODULEMESSAGE,
  '__module__' : 'TeraModuleMessage_pb2'
  # @@protoc_insertion_point(class_scope:opentera.protobuf.TeraModuleMessage)
  })
_sym_db.RegisterMessage(TeraModuleMessage)
_sym_db.RegisterMessage(TeraModuleMessage.Header)


# @@protoc_insertion_point(module_scope)



from opentera.db.Base import db, BaseModel
from opentera.db.models.TeraSite import TeraSite
from opentera.db.models.TeraProject import TeraProject
from flask_sqlalchemy import event

from passlib.hash import bcrypt
import uuid
import datetime
import json


# Generator for jti
def infinite_jti_sequence():
    num = 0
    while True:
        yield num
        num += 1


# Initialize generator, call next(user_jti_generator) to get next sequence number
user_jti_generator = infinite_jti_sequence()


class TeraUser(db.Model, BaseModel):
    __tablename__ = 't_users'
    id_user = db.Column(db.Integer, db.Sequence('id_user_sequence'), primary_key=True, autoincrement=True)
    user_username = db.Column(db.String(50), nullable=False, unique=True)
    user_uuid = db.Column(db.String(36), nullable=False, unique=True)
    user_email = db.Column(db.String, nullable=True)
    user_firstname = db.Column(db.String, nullable=False)
    user_lastname = db.Column(db.String, nullable=False)
    user_password = db.Column(db.String, nullable=False)
    user_enabled = db.Column(db.Boolean, nullable=False)
    user_profile = db.Column(db.String, nullable=False)  # Used to store "extra" informations, if needed.
    user_notes = db.Column(db.String, nullable=True)
    user_lastonline = db.Column(db.TIMESTAMP(timezone=True), nullable=True)
    user_superadmin = db.Column(db.Boolean, nullable=False, default=False)

    # user_sites_access = db.relationship('TeraSiteAccess', cascade="all,delete")
    # user_projects_access = db.relationship("TeraProjectAccess", cascade="all,delete")
    user_user_groups = db.relationship("TeraUserGroup", secondary="t_users_users_groups",
                                       back_populates="user_group_users")
    user_sessions = db.relationship("TeraSession", secondary="t_sessions_users", back_populates="session_users",
                                    passive_deletes=True)

    authenticated = False

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.extend(['authenticated', 'user_password', 'user_user_groups',
                              'user_sessions'])
        if minimal:
            ignore_fields.extend(['user_username', 'user_email', 'user_profile', 'user_notes', 'user_lastonline',
                                  'user_superadmin'])
        rval = super().to_json(ignore_fields=ignore_fields)
        rval['user_name'] = self.get_fullname()
        return rval

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_user': self.id_user, 'user_uuid': self.user_uuid}

    def get_token(self, token_key: str, expiration=3600):
        import time
        import jwt
        import random

        # Creating token with user info
        now = time.time()

        payload = {
            'iat': int(now),
            'exp': int(now) + expiration,
            'iss': 'TeraServer',
            'jti': next(user_jti_generator),
            'user_uuid': self.user_uuid,
            'id_user': self.id_user,
            'user_fullname': self.get_fullname(),
            'user_superadmin': self.user_superadmin
        }

        return jwt.encode(payload, token_key, algorithm='HS256')

    def get_fullname(self):
        return self.user_firstname + ' ' + self.user_lastname

    def is_authenticated(self):
        return self.authenticated

    def is_anonymous(self):
        return False

    def is_active(self):
        return self.user_enabled

    def get_id(self):
        return self.user_uuid

    def update_last_online(self):
        self.user_lastonline = datetime.datetime.now()
        db.session.commit()

    def __str__(self):
        return '<TeraUser ' + str(self.user_username) + ', ' + str(self.user_email) + ' >'

    def __repr__(self):
        return self.__str__()

    def get_sites_roles(self) -> dict:
        sites_roles = {}

        if self.user_superadmin:
            # Super admin - admin role in all sites
            sites = TeraSite.query.all()
            for site in sites:
                sites_roles[site] = {'site_role': 'admin', 'inherited': True}
            return sites_roles

        # Browse all user groups to get roles for those sites
        for user_group in self.user_user_groups:
            user_group_roles = user_group.get_sites_roles()
            for site, site_role in user_group_roles.items():
                if site not in sites_roles:
                    # Site not already present
                    sites_roles[site] = site_role
                else:
                    # Site present - check if we have an "admin" role to overwrite an "user" role
                    if site_role['site_role'] == 'admin':
                        sites_roles[site] = site_role

        return sites_roles

    def get_projects_roles(self) -> dict:
        projects_roles = {}

        if self.user_superadmin:
            # Super admin - admin role in all projects
            projects = TeraProject.query.all()
            for project in projects:
                projects_roles[project] = {'project_role': 'admin', 'inherited': True}
            return projects_roles

        # Browse all user groups to get roles for those projects
        for user_group in self.user_user_groups:
            user_group_roles = user_group.get_projects_roles()
            for project, project_role in user_group_roles.items():
                if project not in projects_roles:
                    # Project not already present
                    projects_roles[project] = project_role
                else:
                    # Project present - check if we have an "admin" role to overwrite an "user" role
                    if project_role['project_role'] == 'admin':
                        projects_roles[project] = project_role

        return projects_roles

    @staticmethod
    def encrypt_password(password):
        return bcrypt.hash(password)

    @staticmethod
    def verify_password(username, password, user=None):
        # Query User with that username
        if user is None:
            user = TeraUser.get_user_by_username(username)
        if user is None:
            print('TeraUser: verify_password - user ' + username + ' not found.')
            return None

        # Check if enabled
        if not user.user_enabled:
            print('TeraUser: verify_password - user ' + username + ' is disabled.')
            return None

        # Check password
        if bcrypt.verify(password, user.user_password):
            user.authenticated = True
            return user

        return None

    @staticmethod
    def get_user_by_username(username):
        return TeraUser.query.filter_by(user_username=username).first()

    @staticmethod
    def get_user_by_uuid(u_uuid):
        user = TeraUser.query.filter_by(user_uuid=u_uuid).first()
        return user

    @staticmethod
    def get_user_by_id(id_user):
        user = TeraUser.query.filter_by(id_user=id_user).first()
        return user

    @staticmethod
    def get_superadmins():
        return TeraUser.query.filter_by(user_superadmin=True).all()

    @classmethod
    def update(cls, id_user: int, values: dict):
        # Remove the password field is present and if empty
        if 'user_password' in values:
            if values['user_password'] == '':
                del values['user_password']
            else:
                # Forcing password to string
                values['user_password'] = TeraUser.encrypt_password(str(values['user_password']))

        # Dumps the user profile if present
        if 'user_profile' in values:
            if isinstance(values['user_profile'], dict):
                values['user_profile'] = json.dumps(values['user_profile'])

        # Prevent changes on UUID
        if 'user_uuid' in values:
            del values['user_uuid']

        super().update(id_user, values)

    @classmethod
    def insert(cls, user):
        # Encrypts password
        # Forcing password to string
        user.user_password = TeraUser.encrypt_password(str(user.user_password))

        # Generate UUID
        user.user_uuid = str(uuid.uuid4())

        # Clear last online field
        user.user_lastonline = None

        # Converts profile from dict, if needed
        if isinstance(user.user_profile, dict):
            user.user_profile = json.dumps(user.user_profile)

        super().insert(user)

    @staticmethod
    def create_defaults(test=False):
        from opentera.db.models.TeraUserGroup import TeraUserGroup
        # Admin
        admin = TeraUser()
        admin.user_enabled = True
        admin.user_firstname = "Super"
        admin.user_lastname = "Admin"
        admin.user_profile = ""
        admin.user_password = TeraUser.encrypt_password("admin")
        admin.user_superadmin = True
        admin.user_username = "admin"
        admin.user_uuid = str(uuid.uuid4())
        db.session.add(admin)

        if test:
            # Site admin
            admin = TeraUser()
            admin.user_enabled = True
            admin.user_firstname = "Site"
            admin.user_lastname = "Admin"
            admin.user_profile = ""
            admin.user_password = TeraUser.encrypt_password("siteadmin")
            admin.user_superadmin = False
            admin.user_username = "siteadmin"
            admin.user_uuid = str(uuid.uuid4())
            # admin.user_user_group = TeraUserGroup.get_user_group_by_group_name("Admins - Default Site")
            db.session.add(admin)

            # Site User
            user = TeraUser()
            user.user_enabled = True
            user.user_firstname = "Site"
            user.user_lastname = "User"
            user.user_profile = ""
            user.user_password = TeraUser.encrypt_password("user")
            user.user_superadmin = False
            user.user_username = "user"
            user.user_uuid = str(uuid.uuid4())
            # user.user_user_group = TeraUserGroup.get_user_group_by_group_name("Users - Project 1")
            db.session.add(user)

            # Site User
            user = TeraUser()
            user.user_enabled = True
            user.user_firstname = "MultiSite"
            user.user_lastname = "User"
            user.user_profile = ""
            user.user_password = TeraUser.encrypt_password("user2")
            user.user_superadmin = False
            user.user_username = "user2"
            user.user_uuid = str(uuid.uuid4())
            # user.user_user_group = TeraUserGroup.get_user_group_by_group_name("Users - Projects 1 & 2")
            db.session.add(user)

            # Project admin
            user = TeraUser()
            user.user_enabled = True
            user.user_firstname = "Project"
            user.user_lastname = "Admin"
            user.user_profile = ""
            user.user_password = TeraUser.encrypt_password("user3")
            user.user_superadmin = False
            user.user_username = "user3"
            user.user_uuid = str(uuid.uuid4())
            # user.user_user_group = TeraUserGroup.get_user_group_by_group_name("Users - Projects 1 & 2")
            db.session.add(user)

            # No access user!
            user = TeraUser()
            user.user_enabled = True
            user.user_firstname = "No Access"
            user.user_lastname = "User!"
            user.user_profile = ""
            user.user_password = TeraUser.encrypt_password("user4")
            user.user_superadmin = False
            user.user_username = "user4"
            user.user_uuid = str(uuid.uuid4())
            # user.user_user_group = TeraUserGroup.get_user_group_by_group_name("Users - Projects 1 & 2")
            db.session.add(user)

        db.session.commit()



from opentera.db.Base import db, BaseModel


class TeraSessionDevices(db.Model, BaseModel):
    __tablename__ = 't_sessions_devices'
    id_session_device = db.Column(db.Integer, db.Sequence('id_session_device'), primary_key=True, autoincrement=True)
    id_session = db.Column(db.Integer, db.ForeignKey('t_sessions.id_session', ondelete='cascade'))
    id_device = db.Column(db.Integer, db.ForeignKey('t_devices.id_device'))

    session_device_session = db.relationship('TeraSession', cascade="delete")
    session_device_device = db.relationship('TeraDevice')



from opentera.db.Base import db, BaseModel
from opentera.db.models.TeraDeviceType import TeraDeviceType
from opentera.db.models.TeraServerSettings import TeraServerSettings

import uuid
import jwt
import time
import datetime


class TeraDevice(db.Model, BaseModel):
    __tablename__ = 't_devices'
    id_device = db.Column(db.Integer, db.Sequence('id_device_sequence'), primary_key=True, autoincrement=True)
    # id_site = db.Column(db.Integer, db.ForeignKey("t_sites.id_site", ondelete='cascade'), nullable=True)
    # id_session_type = db.Column(db.Integer, db.ForeignKey("t_sessions_types.id_session_type",
    #                                                       ondelete='set null'), nullable=True)
    device_uuid = db.Column(db.String(36), nullable=False, unique=True)
    device_name = db.Column(db.String, nullable=False)
    id_device_type = db.Column(db.Integer, db.ForeignKey('t_devices_types.id_device_type', ondelete='cascade'),
                               nullable=False)
    id_device_subtype = db.Column(db.Integer, db.ForeignKey('t_devices_subtypes.id_device_subtype',
                                                            ondelete='set null'), nullable=True)
    device_token = db.Column(db.String, nullable=True, unique=True)
    device_certificate = db.Column(db.String, nullable=True)
    device_enabled = db.Column(db.Boolean, nullable=False, default=False)
    device_onlineable = db.Column(db.Boolean, nullable=False, default=False)
    device_config = db.Column(db.String, nullable=True)
    device_infos = db.Column(db.String, nullable=True)
    device_notes = db.Column(db.String, nullable=True)
    device_lastonline = db.Column(db.TIMESTAMP(timezone=True), nullable=True)

    # device_sites = db.relationship("TeraDeviceSite")
    # device_projects = db.relationship('TeraDeviceProject', cascade='delete')
    device_projects = db.relationship("TeraProject", secondary="t_devices_projects", back_populates="project_devices")
    # device_session_types = db.relationship("TeraSessionTypeDeviceType")
    device_participants = db.relationship("TeraParticipant",  secondary="t_devices_participants",
                                          back_populates="participant_devices", passive_deletes=True)
    device_sessions = db.relationship("TeraSession", secondary="t_sessions_devices", back_populates="session_devices",
                                      passive_deletes=True)
    device_subtype = db.relationship('TeraDeviceSubType')
    device_assets = db.relationship('TeraAsset', passive_deletes=True)

    authenticated = False

    # def __init__(self):
    #     self.secret = TeraServerSettings.get_server_setting_value(TeraServerSettings.ServerDeviceTokenKey)
    #     if self.secret is None:
    #         # Fallback - should not happen
    #         self.secret = 'TeraDeviceSecret'

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields += ['device_projects', 'device_participants', 'device_sessions', 'device_certificate',
                          'device_subtype', 'authenticated', 'device_assets']

        if minimal:
            ignore_fields += ['device_onlineable', 'device_config', 'device_notes',
                              'device_lastonline', 'device_infos',  'device_token']

        device_json = super().to_json(ignore_fields=ignore_fields)

        return device_json

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_device': self.id_device, 'device_uuid': self.device_uuid}

    def from_json(self, json):
        super().from_json(json)

        # Manage device subtype
        if type(self.device_subtype) is dict:
            self.id_device_subtype = self.device_subtype['id_device_subtype']
            self.device_subtype = None

    def is_authenticated(self):
        return self.authenticated

    def is_anonymous(self):
        return False

    def is_active(self):
        return self.device_enabled

    def get_id(self):
        return self.device_uuid

    def is_login_enabled(self):
        return self.device_onlineable

    def create_token(self):
        # Creating token with user info
        payload = {
            'iss': 'TeraServer',
            'device_uuid': self.device_uuid,
            'id_device': self.id_device
        }

        self.device_token = jwt.encode(payload, TeraServerSettings.get_server_setting_value(
            TeraServerSettings.ServerDeviceTokenKey), algorithm='HS256')

        return self.device_token

    def update_last_online(self):
        self.device_lastonline = datetime.datetime.now()
        db.session.commit()

    @staticmethod
    def get_device_by_token(token):
        device = TeraDevice.query.filter_by(device_token=token).first()

        if device:
            # Validate token, key loaded from DB
            data = jwt.decode(token.encode('utf-8'),
                              TeraServerSettings.get_server_setting_value(
                                  TeraServerSettings.ServerDeviceTokenKey), algorithms='HS256')

            # Only validating UUID since other fields can change in database after token is generated.
            if data['device_uuid'] == device.device_uuid:

                # Update last online
                device.update_last_online()
                device.authenticated = True
                return device
            else:
                return None

        return None

    @staticmethod
    def get_device_by_uuid(dev_uuid):
        device = TeraDevice.query.filter_by(device_uuid=dev_uuid).first()
        return device

    @staticmethod
    def get_device_by_name(name):
        return TeraDevice.query.filter_by(device_name=name).first()

    @staticmethod
    def get_device_by_id(device_id):
        return TeraDevice.query.filter_by(id_device=device_id).first()

    @staticmethod
    # Available device = device not assigned to any participant
    def get_available_devices(ignore_disabled=True):
        if ignore_disabled:
            return TeraDevice.query.outerjoin(TeraDevice.device_participants)\
                .filter(TeraDevice.device_participants == None).all()
        else:
            return TeraDevice.query.filter_by(device_enabled=True).outerjoin(TeraDevice.device_participants).\
                filter(TeraDevice.device_participants == None).all()

    @staticmethod
    # Unavailable device = device assigned to at least one participant
    def get_unavailable_devices(ignore_disabled=True):
        if ignore_disabled:
            return TeraDevice.query.join(TeraDevice.device_participants).all()
        else:
            return TeraDevice.query.filter_by(device_enabled=True).join(TeraDevice.device_participants).all()

    @staticmethod
    def create_defaults(test=False):
        if test:
            device = TeraDevice()
            device.device_name = 'Apple Watch #W05P1'
            # Forcing uuid for tests
            device.device_uuid = 'b707e0b2-e649-47e7-a938-2b949c423f73'  # str(uuid.uuid4())
            device.id_device_type = TeraDeviceType.get_device_type_by_key('capteur').id_device_type
            # device.create_token()
            device.device_enabled = True
            device.device_onlineable = True
            # device.device_site = TeraSite.get_site_by_sitename('Default Site')
            # device.device_participants = [TeraParticipant.get_participant_by_id(1)]
            db.session.add(device)

            device2 = TeraDevice()
            device2.device_name = 'Kit Télé #1'
            device2.device_uuid = str(uuid.uuid4())
            device2.id_device_type = TeraDeviceType.get_device_type_by_key('videoconference').id_device_type
            # device2.create_token()
            device2.device_enabled = True
            device2.device_onlineable = True
            # device2.device_sites = [TeraSite.get_site_by_sitename('Default Site')]
            # device2.device_participants = [TeraParticipant.get_participant_by_id(1),
            #                               TeraParticipant.get_participant_by_id(2)]
            db.session.add(device2)

            device3 = TeraDevice()
            device3.device_name = 'Robot A'
            device3.device_uuid = str(uuid.uuid4())
            device3.id_device_type = TeraDeviceType.get_device_type_by_key('robot').id_device_type
            # device3.create_token()
            device3.device_enabled = False
            device3.device_onlineable = True
            # device3.device_sites = [TeraSite.get_site_by_sitename('Default Site')]
            # device3.device_participants = [TeraParticipant.get_participant_by_id(2)]
            db.session.add(device3)

            db.session.commit()

            # Must create token after devices are created since token contains id_device!
            device.create_token()
            device2.create_token()
            device3.create_token()

            db.session.commit()

    @classmethod
    def insert(cls, device):
        # Generate UUID
        device.device_uuid = str(uuid.uuid4())

        # Clear last online field
        device.device_lastonline = None

        # Check for device subtype
        if device.id_device_subtype == 0:
            device.id_device_subtype = None

        super().insert(device)
        # Create token
        device.create_token()
        db.session.commit()

    @classmethod
    def update(cls, update_id: int, values: dict):
        # Check for device subtype
        if 'id_device_subtype' in values:
            if values['id_device_subtype'] == 0:
                values['id_device_subtype'] = None

        # Prevent changes on UUID
        if 'device_uuid' in values:
            del values['device_uuid']

        # Remove object device_subtype
        if 'device_subtype' in values:
            del values['device_subtype']

        super().update(update_id=update_id, values=values)

    @classmethod
    def delete(cls, id_todel):
        super().delete(id_todel)

        # from opentera.db.models.TeraDeviceData import TeraDeviceData
        # TeraDeviceData.delete_files_for_device(id_todel)



from opentera.forms.TeraForm import *
from flask_babel import gettext
from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from opentera.db.models.TeraParticipantGroup import TeraParticipantGroup


class TeraParticipantGroupForm:

    @staticmethod
    def get_participant_group_form(user_access: DBManagerTeraUserAccess, specific_group_id: int = None,
                                   site_id: int = None):
        form = TeraForm("group")

        # Building lists
        projects = []
        if site_id:
            # Specific project id specified
            projects = user_access.query_projects_for_site(site_id=site_id)
        elif not specific_group_id or specific_group_id not in user_access.get_accessible_groups_ids():
            # No specific group id specified
            projects = user_access.get_accessible_projects()
        else:
            # Specific group id specified
            group = TeraParticipantGroup.get_participant_group_by_id(specific_group_id)
            if group:
                projects = user_access.query_projects_for_site(group.participant_group_project.id_site)

        project_list = []
        for project in projects:
            project_list.append(TeraFormValue(value_id=project.id_project, value=project.project_name + ' [' +
                                                                                project.project_site.site_name + ']'))
        # Sections
        section = TeraFormSection("informations", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_participant_group", gettext("Participant Group ID"), "hidden", True))
        section.add_item(TeraFormItem("participant_group_name", gettext("Participant Group Name"), "text", True))
        section.add_item(TeraFormItem("id_project", gettext("Project"), "array", True, item_values=project_list))
        section.add_item(TeraFormItem("project_name", gettext("Project Name"), "hidden"))

        return form.to_dict()



from opentera.db.Base import db, BaseModel
from flask_sqlalchemy import event


class TeraSite(db.Model, BaseModel):
    __tablename__ = 't_sites'
    id_site = db.Column(db.Integer, db.Sequence('id_site_sequence'), primary_key=True, autoincrement=True)
    site_name = db.Column(db.String, nullable=False, unique=True)

    # site_devices = db.relationship("TeraDeviceSite")
    site_projects = db.relationship("TeraProject", cascade="delete", passive_deletes=True)

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['site_projects'])

        return super().to_json(ignore_fields=ignore_fields)

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_site': self.id_site}

    @staticmethod
    def create_defaults(test=False):
        base_site = TeraSite()
        base_site.site_name = 'Default Site'
        TeraSite.insert(base_site)

        if test:
            base_site = TeraSite()
            base_site.site_name = 'Top Secret Site'
            TeraSite.insert(base_site)

    @staticmethod
    def get_site_by_sitename(sitename):
        return TeraSite.query.filter_by(site_name=sitename).first()

    @staticmethod
    def get_site_by_id(site_id: int):
        return TeraSite.query.filter_by(id_site=site_id).first()

    @staticmethod
    def query_data(filter_args):
        if isinstance(filter_args, tuple):
            return TeraSite.query.filter_by(*filter_args).all()
        if isinstance(filter_args, dict):
            return TeraSite.query.filter_by(**filter_args).all()
        return None

    @classmethod
    def delete(cls, id_todel):
        super().delete(id_todel)

        # from opentera.db.models.TeraSession import TeraSession
        # TeraSession.delete_orphaned_sessions()

    @classmethod
    def insert(cls, site):
        # Creates admin and user roles for that site
        super().insert(site)

        from opentera.db.models.TeraServiceRole import TeraServiceRole
        from opentera.db.models.TeraService import TeraService
        opentera_service_id = TeraService.get_openteraserver_service().id_service
        access_role = TeraServiceRole()
        access_role.id_service = opentera_service_id
        access_role.id_site = site.id_site
        access_role.service_role_name = 'admin'
        db.session.add(access_role)

        access_role = TeraServiceRole()
        access_role.id_service = opentera_service_id
        access_role.id_site = site.id_site
        access_role.service_role_name = 'user'
        db.session.add(access_role)

        db.session.commit()

#
# @event.listens_for(TeraSite, 'after_insert')
# def site_inserted(mapper, connection, target):
#     # By default, creates user and admin roles after a site has been added
#     from opentera.db.models.TeraServiceRole import TeraServiceRole
#     from opentera.db.models.TeraService import TeraService
#
#     access_role = TeraServiceRole()
#     access_role.id_service = Globals.opentera_service_id
#     access_role.id_site = target.id_site
#     access_role.service_role_name = 'admin'
#     db.session.add(access_role)
#
#     access_role = TeraServiceRole()
#     access_role.id_service = Globals.opentera_service_id
#     access_role.id_site = target.id_site
#     access_role.service_role_name = 'user'
#     db.session.add(access_role)



from opentera.db.Base import db, BaseModel
from datetime import datetime
import json
import jsonschema


class TeraServiceConfigSpecific(db.Model, BaseModel):
    __tablename__ = 't_services_configs_specifics'
    id_service_config_specific = db.Column(db.Integer, db.Sequence('id_service_config_specific_sequence'),
                                           primary_key=True, autoincrement=True)
    id_service_config = db.Column(db.Integer, db.ForeignKey('t_services_configs.id_service_config', ondelete='cascade'),
                                  nullable=False)
    service_config_specific_id = db.Column(db.String, nullable=False)
    service_config_specific_config = db.Column(db.String, nullable=False)

    service_config_specific_service_config = db.relationship("TeraServiceConfig")

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_config_specific_service_config'])

        if minimal:
            ignore_fields.extend([])

        json_config = super().to_json(ignore_fields=ignore_fields)

        return json_config

    @staticmethod
    def get_service_config_specific_by_id(s_id: int):
        return TeraServiceConfigSpecific.query.filter_by(id_service_config_specific=s_id).first()

    @staticmethod
    def get_service_config_specifics_for_service_config(service_config_id: int):
        return TeraServiceConfigSpecific.query.filter_by(id_service_config=service_config_id).all()

    def get_last_update_datetime(self) -> datetime:
        return datetime.fromtimestamp(self.version_id/1000)

    @staticmethod
    def insert_or_update_specific_config(service_config_id: int, specific_id: str, config: str):
        # Check if we have an existing specific config for that service_config
        configs = TeraServiceConfigSpecific.get_service_config_specifics_for_service_config(service_config_id)
        insertion = True

        if configs:
            for s_config in configs:
                if s_config.service_config_specific_id == specific_id:
                    # Update
                    s_config.service_config_specific_config = config
                    insertion = False
                    break
        if insertion:
            # Insert
            new_specific_config = TeraServiceConfigSpecific()
            new_specific_config.id_service_config = service_config_id
            new_specific_config.service_config_specific_id = specific_id
            new_specific_config.service_config_specific_config = config
            db.session.add(new_specific_config)
        db.session.commit()



from opentera.db.Base import db, BaseModel
import uuid
from enum import Enum, unique
from sqlalchemy import or_


@unique
class AssetType(Enum):
    RAW_FILE = 1
    RAW_DATA = 2
    RESULT_FILE = 3
    PROCESSED_DATA = 4
    REPORT = 5

    def describe(self):
        return self.name, self.value


class TeraAsset(db.Model, BaseModel):
    __tablename__ = 't_assets'
    id_asset = db.Column(db.Integer, db.Sequence('id_asset_sequence'), primary_key=True, autoincrement=True)
    id_session = db.Column(db.Integer, db.ForeignKey("t_sessions.id_session", ondelete='cascade'), nullable=False)
    id_device = db.Column(db.Integer, db.ForeignKey("t_devices.id_device"), nullable=True)
    id_participant = db.Column(db.Integer, db.ForeignKey("t_participants.id_participant"),
                               nullable=True)
    id_user = db.Column(db.Integer, db.ForeignKey("t_users.id_user"), nullable=True)
    # Put a description of the asset here
    asset_name = db.Column(db.String, nullable=False)

    asset_uuid = db.Column(db.String(36), nullable=False, unique=True)
    asset_service_uuid = db.Column(db.String(36), nullable=False)
    asset_type = db.Column(db.Integer, nullable=False)

    asset_session = db.relationship("TeraSession")
    asset_device = db.relationship("TeraDevice")
    asset_user = db.relationship("TeraUser")
    asset_participant = db.relationship("TeraParticipant")

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['asset_session', 'asset_device', 'asset_user', 'asset_participant'])

        return super().to_json(ignore_fields=ignore_fields)

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_asset': self.id_asset, 'asset_uuid': self.asset_uuid}

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraSession import TeraSession
            from opentera.db.models.TeraDevice import TeraDevice
            from opentera.db.models.TeraParticipant import TeraParticipant
            from opentera.db.models.TeraUser import TeraUser

            session2 = TeraSession.get_session_by_name("Séance #2")
            session3 = TeraSession.get_session_by_name("Séance #3")

            for i in range(3):
                new_asset = TeraAsset()
                new_asset.asset_name = "Asset #" + str(i)
                new_asset.asset_session = session2
                new_asset.asset_uuid = str(uuid.uuid4())
                new_asset.asset_service_uuid = '00000000-0000-0000-0000-000000000001'
                new_asset.asset_type = AssetType.RAW_FILE.value
                if i == 0:
                    new_asset.id_participant = TeraParticipant.get_participant_by_name('Participant #1').id_participant
                if i == 1:
                    new_asset.id_user = TeraUser.get_user_by_id(1).id_user
                db.session.add(new_asset)

            asset_device = TeraDevice.get_device_by_name('Apple Watch #W05P1')
            new_asset = TeraAsset()
            new_asset.asset_name = "Device Asset"
            new_asset.asset_session = session3
            new_asset.asset_device = asset_device
            new_asset.asset_uuid = str(uuid.uuid4())
            new_asset.asset_service_uuid = '00000000-0000-0000-0000-000000000001'
            new_asset.asset_type = AssetType.PROCESSED_DATA.value
            db.session.add(new_asset)

            db.session.commit()

    @staticmethod
    def get_asset_by_id(asset_id: int):
        return TeraAsset.query.filter_by(id_asset=asset_id).first()

    @staticmethod
    def get_assets_for_device(device_id: int):
        # return TeraAsset.query.filter_by(id_device=device_id).all()
        from opentera.db.models.TeraSession import TeraSession
        return TeraAsset.query.join(TeraSession).filter(or_(TeraSession.session_devices.any(
            id_device=device_id), TeraAsset.id_device == device_id)).all()

    @staticmethod
    def get_assets_for_user(user_id: int):
        from opentera.db.models.TeraSession import TeraSession
        return TeraAsset.query.join(TeraSession).filter(or_(TeraSession.session_users.any(
            id_user=user_id), TeraAsset.id_user == user_id)).all()

    @staticmethod
    def get_assets_for_session(session_id: int):
        return TeraAsset.query.filter_by(id_session=session_id).all()

    @staticmethod
    def get_assets_for_participant(part_id: int):
        from opentera.db.models.TeraSession import TeraSession
        return TeraAsset.query.join(TeraSession).filter(or_(TeraSession.session_participants.any(
            id_participant=part_id), TeraAsset.id_participant == part_id)).all()

    @staticmethod
    def get_assets_for_service(service_uuid: str):
        return TeraAsset.query.filter_by(service_uuid=service_uuid).all()

    @classmethod
    def insert(cls, asset):
        # Generate UUID
        asset.asset_uuid = str(uuid.uuid4())

        super().insert(asset)



from opentera.db.Base import db, BaseModel


class TeraServiceSite(db.Model, BaseModel):
    __tablename__ = 't_services_sites'
    id_service_site = db.Column(db.Integer, db.Sequence('id_service_site_sequence'), primary_key=True,
                                autoincrement=True)
    id_service = db.Column(db.Integer, db.ForeignKey('t_services.id_service', ondelete='cascade'), nullable=False)
    id_site = db.Column(db.Integer, db.ForeignKey('t_sites.id_site', ondelete='cascade'), nullable=False)

    service_site_service = db.relationship("TeraService")
    service_site_site = db.relationship("TeraSite")

    def __init__(self):
        pass

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_site_service', 'service_site_site'])

        if minimal:
            ignore_fields.extend([])

        return super().to_json(ignore_fields=ignore_fields)

    @staticmethod
    def get_services_for_site(id_site: int):
        return TeraServiceSite.query.filter_by(id_site=id_site).all()

    @staticmethod
    def get_sites_for_service(id_service: int):
        return TeraServiceSite.query.filter_by(id_service=id_service).all()

    @staticmethod
    def get_service_site_by_id(service_site_id: int):
        return TeraServiceSite.query.filter_by(id_service_site=service_site_id).first()

    @staticmethod
    def get_service_site_for_service_site(site_id: int, service_id: int):
        return TeraServiceSite.query.filter_by(id_site=site_id, id_service=service_id).first()

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraService import TeraService
            from opentera.db.models.TeraSite import TeraSite

            site1 = TeraSite.get_site_by_sitename('Default Site')
            site2 = TeraSite.get_site_by_sitename('Top Secret Site')

            service_bureau = TeraService.get_service_by_key('BureauActif')

            service_site = TeraServiceSite()
            service_site.id_site = site1.id_site
            service_site.id_service = service_bureau.id_service
            db.session.add(service_site)

            service_site = TeraServiceSite()
            service_site.id_site = site2.id_site
            service_site.id_service = service_bureau.id_service
            db.session.add(service_site)

            db.session.commit()

    @staticmethod
    def delete_with_ids(service_id: int, site_id: int):
        delete_obj = TeraServiceSite.query.filter_by(id_service=service_id, id_site=site_id).first()
        if delete_obj:
            db.session.delete(delete_obj)
            db.session.commit()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: LeaveSessionEvent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='LeaveSessionEvent.proto',
  package='opentera.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x17LeaveSessionEvent.proto\x12\x11opentera.protobuf\"\x8d\x01\n\x11LeaveSessionEvent\x12\x14\n\x0csession_uuid\x18\x01 \x01(\t\x12\x14\n\x0cservice_uuid\x18\x02 \x01(\t\x12\x1c\n\x14leaving_participants\x18\x03 \x03(\t\x12\x15\n\rleaving_users\x18\x04 \x03(\t\x12\x17\n\x0fleaving_devices\x18\x05 \x03(\tb\x06proto3'
)




_LEAVESESSIONEVENT = _descriptor.Descriptor(
  name='LeaveSessionEvent',
  full_name='opentera.protobuf.LeaveSessionEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_uuid', full_name='opentera.protobuf.LeaveSessionEvent.session_uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_uuid', full_name='opentera.protobuf.LeaveSessionEvent.service_uuid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='leaving_participants', full_name='opentera.protobuf.LeaveSessionEvent.leaving_participants', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='leaving_users', full_name='opentera.protobuf.LeaveSessionEvent.leaving_users', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='leaving_devices', full_name='opentera.protobuf.LeaveSessionEvent.leaving_devices', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=47,
  serialized_end=188,
)

DESCRIPTOR.message_types_by_name['LeaveSessionEvent'] = _LEAVESESSIONEVENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LeaveSessionEvent = _reflection.GeneratedProtocolMessageType('LeaveSessionEvent', (_message.Message,), {
  'DESCRIPTOR' : _LEAVESESSIONEVENT,
  '__module__' : 'LeaveSessionEvent_pb2'
  # @@protoc_insertion_point(class_scope:opentera.protobuf.LeaveSessionEvent)
  })
_sym_db.RegisterMessage(LeaveSessionEvent)


# @@protoc_insertion_point(module_scope)



from opentera.db.Base import db, BaseModel
from datetime import datetime
import json
import jsonschema


class TeraServiceConfig(db.Model, BaseModel):
    __tablename__ = 't_services_configs'
    id_service_config = db.Column(db.Integer, db.Sequence('id_service_config_sequence'), primary_key=True,
                                  autoincrement=True)
    id_service = db.Column(db.Integer, db.ForeignKey('t_services.id_service', ondelete='cascade'), nullable=False)
    id_user = db.Column(db.Integer, db.ForeignKey('t_users.id_user', ondelete='cascade'), nullable=True)
    id_participant = db.Column(db.Integer, db.ForeignKey('t_participants.id_participant', ondelete='cascade'),
                               nullable=True)
    id_device = db.Column(db.Integer, db.ForeignKey('t_devices.id_device', ondelete='cascade'), nullable=True)
    service_config_config = db.Column(db.String, nullable=False, default='{}')

    service_config_service = db.relationship("TeraService")
    service_config_user = db.relationship("TeraUser")
    service_config_participant = db.relationship('TeraParticipant')
    service_config_device = db.relationship('TeraDevice')
    service_config_specifics = db.relationship('TeraServiceConfigSpecific', cascade='delete')

    def to_json(self, ignore_fields=None, minimal=False, specific_id=None):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_config_service', 'service_config_user', 'service_config_participant',
                              'service_config_device', 'service_config_specifics'])

        if minimal:
            ignore_fields.extend([])

        json_config = super().to_json(ignore_fields=ignore_fields)

        if not minimal:
            json_config['service_config_name'] = self.service_config_service.service_name
            if self.version_id:  # No version ID can occur on temporary objects
                json_config['service_config_last_update_time'] = self.get_last_update_datetime().isoformat()
            else:
                json_config['service_config_last_update_time'] = None

        # Filter null ids
        if not self.id_user:
            del json_config['id_user']

        if not self.id_participant:
            del json_config['id_participant']

        if not self.id_device:
            del json_config['id_device']

        # Get correct config
        if specific_id and self.has_specific_config_for_specific_id(specific_id=specific_id):
            json_config['service_config_config'] = self.get_config_for_specific_id(specific_id=specific_id)
            json_config['service_config_specific_id'] = specific_id
        else:
            json_config['service_config_config'] = self.get_global_config()
            json_config['service_config_specific_id'] = None

        return json_config

    def from_json(self, json_values: dict):
        super().from_json(json=json_values, ignore_fields=['service_config_config', 'id_specific'])

        if 'service_config_config' in json_values:
            if 'id_specific' not in json_values:
                # Specifc config can't be handled here, since id_service_config might be equal to 0...
                #     self.set_config_for_specific_id(json_values['id_specific'], json_values['service_config_config'])
                # else:
                self.set_global_config(json_values['service_config_config'])

    @staticmethod
    def get_service_config_by_id(s_id: int):
        return TeraServiceConfig.query.filter_by(id_service_config=s_id).first()

    @staticmethod
    def get_service_config_for_service(service_id: int):
        return TeraServiceConfig.query.filter_by(id_service=service_id).all()

    @staticmethod
    def get_service_config_for_service_for_user(service_id: int, user_id: int):
        return TeraServiceConfig.query.filter_by(id_service=service_id, id_user=user_id).first()

    @staticmethod
    def get_service_config_for_service_for_participant(service_id: int, participant_id: int):
        return TeraServiceConfig.query.filter_by(id_service=service_id, id_participant=participant_id).first()

    @staticmethod
    def get_service_config_for_service_for_device(service_id: int, device_id: int):
        return TeraServiceConfig.query.filter_by(id_service=service_id, id_device=device_id).first()

    def has_specific_config_for_specific_id(self, specific_id: int) -> bool:
        for specific in self.service_config_specifics:
            if specific.service_config_specific_id == specific_id:
                return True
        return False

    def get_last_update_datetime(self) -> datetime:
        return datetime.fromtimestamp(self.version_id/1000)

    def config_is_valid(self) -> bool:
        if self.service_config_config is None:
            return True  # Will use default value

        # Check if the config is a correct json format
        if isinstance(self.service_config_config, str):
            try:
                json.loads(self.service_config_config)
            except ValueError as err:
                raise err

        return True

    def get_global_config(self) -> dict:
        if self.service_config_config is None:
            return dict()

        try:
            config_json = json.loads(self.service_config_config)
        except ValueError as err:
            raise err
        return config_json

    def get_config_for_specific_id(self, specific_id: str) -> dict:
        if self.service_config_config is None:
            return dict()

        try:
            config_json = json.loads(self.service_config_config)
        except ValueError as err:
            raise err

        # No specific config at all - return global config
        if self.service_config_specifics is None or len(self.service_config_specifics) == 0:
            return config_json

        # Check if we have a specific config
        for specific in self.service_config_specifics:
            if specific.service_config_specific_id == specific_id:
                # Complete or overwrite values in current config
                try:
                    specific_config = json.loads(specific.service_config_specific_config)
                except ValueError as err:
                    raise err
                for specific_key, specific_value in specific_config.items():
                    config_json[specific_key] = specific_value

        return config_json

    def set_config_for_specific_id(self, specific_id: str, config: str):
        if isinstance(config, dict):
            try:
                config = json.dumps(config)
            except ValueError as err:
                raise err

        #  Set or update specific config
        from opentera.db.models.TeraServiceConfigSpecific import TeraServiceConfigSpecific
        TeraServiceConfigSpecific.insert_or_update_specific_config(service_config_id=self.id_service_config,
                                                                   specific_id=specific_id, config=config)

    def set_global_config(self, config: str):
        if isinstance(config, dict):
            try:
                config = json.dumps(config)
            except ValueError as err:
                raise err

        self.service_config_config = config

    @staticmethod
    def get_empty_config() -> dict:
        # return {'Globals': {}, 'Specifics': {}}
        return dict()

    def get_specific_ids_list(self) -> list:
        return [specific_config.service_config_specific_id for specific_config in self.service_config_specifics]

    @staticmethod
    def create_defaults(test=False):
        if test:
            from .TeraUser import TeraUser
            from .TeraDevice import TeraDevice
            from .TeraService import TeraService
            from .TeraParticipant import TeraParticipant
            from .TeraServiceConfigSpecific import TeraServiceConfigSpecific

            new_config = TeraServiceConfig()
            new_config.id_user = TeraUser.get_user_by_id(1).id_user
            new_config.id_service = TeraService.get_openteraserver_service().id_service
            new_config.service_config_config = '{"notification_sounds": true}'
            db.session.add(new_config)

            new_config = TeraServiceConfig()
            new_config.id_participant = TeraParticipant.get_participant_by_name('Participant #1').id_participant
            new_config.id_service = TeraService.get_service_by_key('VideoRehabService').id_service
            new_config.service_config_config = '{"default_muted": false, "view_local": true}'
            db.session.add(new_config)
            db.session.commit()

            new_specific_config = TeraServiceConfigSpecific()
            new_specific_config.id_service_config = new_config.id_service_config
            new_specific_config.service_config_specific_id = 'pc-001'
            new_specific_config.service_config_specific_config = '{"default_muted": true}'
            db.session.add(new_specific_config)

            new_specific_config = TeraServiceConfigSpecific()
            new_specific_config.id_service_config = new_config.id_service_config
            new_specific_config.service_config_specific_id = 'pc-002'
            new_specific_config.service_config_specific_config = '{"default_muted": true, "view_local": false}'
            db.session.add(new_specific_config)

            new_config = TeraServiceConfig()
            new_config.id_device = TeraDevice.get_device_by_name('Apple Watch #W05P1').id_device
            new_config.id_service = TeraService.get_service_by_key('VideoRehabService').id_service
            new_config.service_config_config = '{"delete_from_device": true}'
            db.session.add(new_config)

            db.session.commit()

    @classmethod
    def update(cls, update_id: int, values: dict):
        # Validate that the config we have is valid
        if 'service_config_config' in values:
            if 'id_specific' in values:
                # We have a specific config...
                config = TeraServiceConfig.get_service_config_by_id(values['id_service_config'])
                if not config:
                    return False

                config.set_config_for_specific_id(specific_id=values['id_specific'],
                                                  config=values['service_config_config'])
            else:
                # We have a global config...
                config = TeraServiceConfig.get_service_config_by_id(values['id_service_config'])
                if not config:
                    return False

                config.set_global_config(config=values['service_config_config'])

            if not config.config_is_valid():
                return False

            # Already updated - remove from list
            del values['service_config_config']

        super().update(update_id=update_id, values=values)
        return True

    @classmethod
    def insert(cls, db_object):
        # Validate that the config we have is valid
        try:
            config_valid = db_object.config_is_valid()
        except (ValueError, jsonschema.exceptions.ValidationError) as err:
            raise err

        if not config_valid:
            return False

        super().insert(db_object)
        return True

